from typing import Tuple

import numpy as np
import pyarrow.compute as pc

from .differential_correction import OrbitDeterminationObservations
from .fitted_orbits import FittedOrbitMembers


def remove_lowest_probability_observation(
    orbit_members: FittedOrbitMembers, observations: OrbitDeterminationObservations
) -> Tuple[str, OrbitDeterminationObservations]:
    """
    Remove the observation with the worst residual from the observations. The probability is defined to be
    the probability of drawing a more extreme residual than the one observed. If multiple observations
    have the same probability, then the observation with the highest squared residual value is removed.

    Parameters
    ----------
    orbit_members : FittedOrbitMembers
        The orbit members that contain the residuals with respect to the observations.
    observations : OrbitDeterminationObservations
        The observations to remove the worst residual from.

    Returns
    -------
    obs_id : str
        The ID of the observation that was removed.
    filtered_observations : OrbitDeterminationObservations
        The observations with the worst residual removed.
    """
    assert (
        len(pc.unique(orbit_members.orbit_id)) == 1
    ), "Orbit members must only contain one orbit"
    assert pc.all(
        pc.is_in(orbit_members.obs_id, observations.id)
    ).as_py(), "Observations must contain all orbit member observations"

    # Find the worst outlier (the observation that has the lowest probability of
    # drawing a more extreme residual than the one observed)
    worst_outlier = orbit_members.apply_mask(
        pc.equal(
            orbit_members.residuals.probability,
            pc.min(orbit_members.residuals.probability),
        )
    )

    if len(worst_outlier) > 1:
        # If there are multiple worst outliers (which would be quite unlikely),
        # then remove the outlier with the highest squared residual value
        index = np.nansum(worst_outlier.residuals.to_array() ** 2, axis=1).argmax()
        worst_outlier = worst_outlier.take([index])

    # Grab the observation ID of the worst outlier
    obs_id = worst_outlier.obs_id[0].as_py()

    # Grab the surviving observation IDs
    obs_ids = orbit_members.apply_mask(
        pc.invert(pc.equal(orbit_members.obs_id, obs_id))
    ).obs_id

    return obs_id, observations.apply_mask(pc.is_in(observations.id, obs_ids))



from __future__ import annotations

from typing import Iterator

import healpy
import numpy as np
import numpy.typing as npt
import pyarrow
import pyarrow.compute
import quivr as qv
from quivr.validators import and_, ge, le

from ..time import Timestamp
from .exposures import Exposures


class PointSourceDetections(qv.Table):
    """
    PointSourceDetections is a table of data about point source detections.

    """

    id = qv.LargeStringColumn()

    exposure_id = qv.LargeStringColumn(nullable=True)

    # Some, but not all, point source data may include times for
    # individual observations within an exposure.
    time = Timestamp.as_column(nullable=True)

    # Should this be a SphericalCoordinates instead?

    ra = qv.Float64Column(validator=and_(ge(0), le(360)))
    ra_sigma = qv.Float64Column(nullable=True)

    dec = qv.Float64Column(validator=and_(ge(-90), le(90)))
    dec_sigma = qv.Float64Column(nullable=True)

    mag = qv.Float64Column(validator=and_(ge(-10), le(30)))
    mag_sigma = qv.Float64Column(nullable=True)

    def group_by_exposure(self) -> Iterator["PointSourceDetections"]:
        """
        Returns an iterator of PointSourceDetections, each grouped by exposure_id.
        """
        # Gather unique exposure IDs
        exposure_ids = self.exposure_id.unique()
        sorted = self.table.sort_by("exposure_id")
        for exposure_id in exposure_ids:
            mask = pyarrow.compute.equal(sorted.column("exposure_id"), exposure_id)
            table = sorted.filter(mask)
            yield PointSourceDetections(table)

    def healpixels(self, nside: int, nest: bool = True) -> npt.NDArray[np.int64]:
        """
        Returns an array of healpixels for each observation.
        """
        return healpy.ang2pix(nside, self.ra, self.dec, nest=nest, lonlat=True)

    def group_by_healpixel(
        self, nside: int, nest: bool = True
    ) -> Iterator[tuple[int, PointSourceDetections]]:
        """
        Returns an iterator of PointSourceDetections, each grouped by healpixel.
        """
        # Gather unique healpixels
        healpixels = self.healpixels(nside, nest)
        unique_healpixels = np.unique(healpixels)
        for healpixel in unique_healpixels:
            mask = pyarrow.compute.equal(healpixels, healpixel)
            yield (healpixel, self.apply_mask(mask))

    def link_to_exposures(
        self, exposures: Exposures
    ) -> qv.Linkage[PointSourceDetections, Exposures]:
        """
        Links the detections to the exposures.
        """
        return qv.Linkage(self, exposures, self.exposure_id, exposures.id)



# flake8: noqa: F401

# TODO: associations - should it be in the API? its useful for test helpers, but kind of niche
from .associations import Associations
from .detections import PointSourceDetections
from .exposures import Exposures

__all__ = [
    "Associations",
    "Exposures",
    "PointSourceDetections",
]



from typing import Tuple

import jax.numpy as jnp
from jax import config, jit, lax

config.update("jax_enable_x64", True)

STUMPFF_TYPES = Tuple[
    jnp.float64, jnp.float64, jnp.float64, jnp.float64, jnp.float64, jnp.float64
]


@jit
def _positive_psi(psi: jnp.float64) -> STUMPFF_TYPES:
    # Equation 6.9.15 in Danby (1992) [1]
    sqrt_psi = jnp.sqrt(psi)
    c0 = jnp.cos(sqrt_psi)
    c1 = jnp.sin(sqrt_psi) / sqrt_psi

    # Equation 6.9.16 in Danby (1992) [1]
    # states the recursion relation for higher
    # order Stumpff functions
    c2 = (1.0 - c0) / psi
    c3 = (1.0 - c1) / psi
    c4 = (1 / 2.0 - c2) / psi
    c5 = (1 / 6.0 - c3) / psi

    return c0, c1, c2, c3, c4, c5


@jit
def _negative_psi(psi: jnp.float64) -> STUMPFF_TYPES:
    # Equation 6.9.15 in Danby (1992) [1]
    sqrt_npsi = jnp.sqrt(-psi)
    c0 = jnp.cosh(sqrt_npsi)
    c1 = jnp.sinh(sqrt_npsi) / sqrt_npsi

    # Equation 6.9.16 in Danby (1992) [1]
    # states the recursion relation for higher
    # order Stumpff functions
    c2 = (1.0 - c0) / psi
    c3 = (1.0 - c1) / psi
    c4 = (1 / 2.0 - c2) / psi
    c5 = (1 / 6.0 - c3) / psi

    return c0, c1, c2, c3, c4, c5


@jit
def _null_psi(psi: jnp.float64) -> STUMPFF_TYPES:
    # Equation 6.9.14 in Danby (1992) [1]
    c0 = 1.0
    c1 = 1.0
    c2 = 1 / 2.0
    c3 = 1 / 6.0
    c4 = 1 / 24.0
    c5 = 1 / 120.0

    return c0, c1, c2, c3, c4, c5


@jit
def calc_stumpff(psi: jnp.float64) -> STUMPFF_TYPES:
    """
    Calculate the first 6 Stumpff functions for variable psi.

    Parameters
    ----------
    psi : float
        Dimensionless parameter at which to evaluate the Stumpff functions (equivalent to alpha * chi**2).

    Returns
    -------
    c0, c1, c2, c3, c4, c5 : 6 x float
        First six Stumpff functions.

    References
    ----------
    [1] Danby, J. M. A. (1992). Fundamentals of Celestial Mechanics. 2nd ed.,
        William-Bell, Inc. ISBN-13: 978-0943396200
        Notes: of particular interest is Danby's fantastic chapter on universal variables (6.9)
    """
    c0, c1, c2, c3, c4, c5 = lax.cond(
        psi > 0.0,
        _positive_psi,
        lambda psi: lax.cond(psi < 0.0, _negative_psi, _null_psi, psi),
        psi,
    )
    return c0, c1, c2, c3, c4, c5



import warnings

import pyarrow as pa
import quivr as qv

from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.spherical import SphericalCoordinates
from ..observers.observers import Observers


class Ephemeris(qv.Table):

    orbit_id = qv.LargeStringColumn()
    object_id = qv.LargeStringColumn(nullable=True)
    coordinates = SphericalCoordinates.as_column()

    # The coordinates as observed by the observer will be the result of
    # light emitted or reflected from the object at the time of the observation.
    # Light, however, has a finite speed and so the object's observed cooordinates
    # will be different from its actual geometric coordinates at the time of observation.
    # Aberrated coordinates are coordinates that account for the light travel time
    # from the time of emission/reflection to the time of observation
    alpha = qv.Float64Column(nullable=True)
    light_time = qv.Float64Column(nullable=True)
    aberrated_coordinates = CartesianCoordinates.as_column(nullable=True)

    def link_to_observers(
        self, observers: Observers, precision="ns"
    ) -> qv.MultiKeyLinkage["Ephemeris", Observers]:
        """
        Link these ephemerides back to the observers that generated them. This is useful if
        you want or need to use the observer's position as part of any computation for
        any given set of ephemerides.

        Not all propagators will return ephemerides exactly at the time of the input observers.
        As an example, PYOORB stores times as a single MJD, when converting from two integers to
        this singular float there will be a loss of precision. To mitigate this, the user may
        optionally define the precision to which would like to link back to observers. Times
        for both the ephemerides and observers will be rounded to this precision before linking.

        Parameters
        ----------
        observers : `~adam_core.observers.observers.Observers` (N)
            Observers that generated the ephemerides.
        precision : str, optional
            Precision to which to link back to observers, by default "ns".

        Returns
        -------
        `~qv.MultiKeyLinkage[
                `~adam_core.orbits.ephemeris.Ephemeris`,
                `~adam_core.observers.observersObservers
        ]`
            Linkage between ephemerides and observers.
        """
        if self.coordinates.time.scale != observers.coordinates.time.scale:
            observers = observers.set_column(
                "coordinates.time",
                observers.coordinates.time.rescale(self.coordinates.time.scale),
            )

        rounded = self.coordinates.time.rounded(precision)
        observers_rounded = observers.coordinates.time.rounded(precision)

        left_keys = {
            "days": rounded.days,
            "nanos": rounded.nanos,
            "observatory_code": self.coordinates.origin.code,
        }
        right_keys = {
            "days": observers_rounded.days,
            "nanos": observers_rounded.nanos,
            "observatory_code": observers.code,
        }
        linkage = qv.MultiKeyLinkage(self, observers, left_keys, right_keys)

        # Check to make sure we have the correct number of linkages by calculating the number
        # of unique observers in the table and checking to see if that matches the number of
        # unique keys in the linkage
        observers_table = pa.table(
            [
                observers.code,
                observers.coordinates.time.days,
                observers.coordinates.time.nanos,
            ],
            names=["observatory_code", "days", "nanos"],
        )
        unique_observers = observers_table.group_by(
            ["observatory_code", "days", "nanos"]
        ).aggregate([])

        expected_length = len(unique_observers)
        actual_length = len(linkage.all_unique_values)
        if expected_length != actual_length:
            warnings.warn(
                "The number of unique keys in the linkage does not match the number"
                " of unique observers. Linkage precision may be too low."
                f"Expected {expected_length} unique keys, got {actual_length}.",
                UserWarning,
            )
        return linkage



# flake8: noqa: F401
def test_import_observers():
    from adam_core.observers import Observers, get_observer_state


def test_import_orbits():
    from adam_core.orbits import Ephemeris, Orbits, VariantOrbits
    from adam_core.orbits.query import query_horizons, query_sbdb


def test_import_dynamics():
    from adam_core.dynamics import propagate_2body


def test_import_coordinates():
    from adam_core.coordinates import (
        CartesianCoordinates,
        CometaryCoordinates,
        CoordinateCovariances,
        KeplerianCoordinates,
        Origin,
        OriginCodes,
        SphericalCoordinates,
        transform_coordinates,
    )


def test_import_observations():
    from adam_core.observations import Associations, Exposures, PointSourceDetections


def test_import_utils():
    from adam_core.utils import get_perturber_state, setup_SPICE


def test_import_time():
    from adam_core.time import Timestamp



import os

import numpy as np
import pytest

from ...constants import KM_P_AU, S_P_DAY
from ...coordinates.cartesian import CartesianCoordinates
from ...coordinates.origin import OriginCodes
from ...coordinates.residuals import Residuals
from ...observers import get_observer_state
from ...time import Timestamp
from ...utils import get_perturber_state

DATA_DIR = os.path.join(os.path.dirname(__file__), "testdata")


def test_get_observer_state_OriginCodes():
    # Test that get_observer_state works using OriginCodes instead of MPC
    # observatory codes
    codes = [
        OriginCodes.SUN,
        OriginCodes.MERCURY,
        OriginCodes.VENUS,
        OriginCodes.EARTH,
        OriginCodes.MARS_BARYCENTER,
        OriginCodes.JUPITER_BARYCENTER,
    ]
    for code in codes:

        times = Timestamp.from_mjd(np.arange(59000, 60000), scale="tdb")
        perturber_state = get_perturber_state(
            code, times, frame="ecliptic", origin=OriginCodes.SOLAR_SYSTEM_BARYCENTER
        )
        observer_state = get_observer_state(
            code, times, frame="ecliptic", origin=OriginCodes.SOLAR_SYSTEM_BARYCENTER
        )
        assert perturber_state == observer_state


expected_precision = {
    # Units of m and mm/s
    "X05": {"r": 15, "v": 2},
    "I41": {"r": 10, "v": 2},
    "W84": {"r": 10, "v": 2},
    "F51": {"r": 15, "v": 2},
    "000": {"r": 10, "v": 2},
    "500": {"r": 0.10, "v": 0.01},
}


@pytest.mark.parametrize("code", expected_precision.keys())
@pytest.mark.parametrize("origin", ["sun", "ssb"])
def test_get_observer_state(code, origin):
    # Test that we can get the observer state for X05 to within 15 m and 2 mm/s
    # using the MPC extended observatory codes file and SPICE kernels
    states_expected = CartesianCoordinates.from_parquet(
        os.path.join(DATA_DIR, f"{code}_{origin}.parquet"),
    )

    if origin == "sun":
        origin_code = OriginCodes.SUN
    else:
        origin_code = OriginCodes.SOLAR_SYSTEM_BARYCENTER

    # Get the observer state using adam_core
    states = get_observer_state(
        code,
        states_expected.time,
        frame=states_expected.frame,
        origin=origin_code,
    )

    # Calculate residuals and extract values
    residuals = Residuals.calculate(states, states_expected)
    residual_values = np.stack(residuals.values.to_numpy(zero_copy_only=False))

    # Calculate the offset in position in meters
    r_offset = np.linalg.norm(residual_values[:, :3], axis=1) * KM_P_AU * 1000
    # Calculate the offset in velocity in mm/s
    v_offset = (
        np.linalg.norm(residual_values[:, 3:], axis=1) * KM_P_AU / S_P_DAY * 1000000
    )
    np.testing.assert_array_less(r_offset, expected_precision[code]["r"])
    np.testing.assert_array_less(v_offset, expected_precision[code]["v"])


def test_get_observer_state_raises():
    # Test that when we ask for a space-based observatory we raise an error
    code = "C51"
    with pytest.raises(ValueError):
        # Get the observer state using adam_core
        get_observer_state(
            code,
            Timestamp.from_mjd([59000, 60000], scale="tdb"),
        )



import jax.numpy as jnp
import numpy as np
from jax import config, jit, vmap

from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.covariances import (
    CoordinateCovariances,
    transform_covariances_jacobian,
)
from ..coordinates.origin import Origin
from ..orbits.orbits import Orbits
from ..time import Timestamp
from .lagrange import apply_lagrange_coefficients, calc_lagrange_coefficients

config.update("jax_enable_x64", True)


@jit
def _propagate_2body(
    orbit: jnp.ndarray,
    t0: float,
    t1: float,
    mu: float,
    max_iter: int = 1000,
    tol: float = 1e-14,
) -> jnp.ndarray:
    """
    Propagate an orbit from t0 to t1.

    Parameters
    ----------
    orbit : `~jax.numpy.ndarray` (6)
        Cartesian orbit with position in units of au and velocity in units of au per day.
    t0 : float (1)
        Epoch in MJD at which the orbit are defined.
    t1 : float (N)
        Epochs to which to propagate the given orbit.
    mu : float (1)
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will return the value of the universal anomaly at the last iteration.
    tol : float, optional
        Numerical tolerance to which to compute universal anomaly using the Newtown-Raphson
        method.

    Returns
    -------
    orbits : `~jax.numpy.ndarray` (N, 6)
        Orbit propagated to each MJD with position in units of au and velocity in units
        of au per day.
    """
    r = orbit[0:3]
    v = orbit[3:6]
    dt = t1 - t0

    lagrange_coeffs, stumpff_coeffs, chi = calc_lagrange_coefficients(
        r, v, dt, mu=mu, max_iter=max_iter, tol=tol
    )
    r_new, v_new = apply_lagrange_coefficients(r, v, *lagrange_coeffs)

    return jnp.array([r_new[0], r_new[1], r_new[2], v_new[0], v_new[1], v_new[2]])


# Vectorization Map: _propagate_2body
_propagate_2body_vmap = jit(
    vmap(_propagate_2body, in_axes=(0, 0, 0, 0, None, None), out_axes=(0))
)


def propagate_2body(
    orbits: Orbits,
    times: Timestamp,
    max_iter: int = 1000,
    tol: float = 1e-14,
) -> Orbits:
    """
    Propagate orbits using the 2-body universal anomaly formalism.

    Parameters
    ----------
    orbits : `~jax.numpy.ndarray` (N, 6)
        Cartesian orbits with position in units of au and velocity in units of au per day.
    times : Timestamp (M)
        Epochs to which to propagate each orbit. If a single epoch is given, all orbits are propagated to this
        epoch. If multiple epochs are given, then each orbit to will be propagated to each epoch.
    max_iter : int, optional
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will return the value of the universal anomaly at the last iteration.
    tol : float, optional
        Numerical tolerance to which to compute universal anomaly using the Newtown-Raphson
        method.

    Returns
    -------
    orbits : `~adam_core.orbits.orbits.Orbits` (N*M)
        Orbits propagated to each MJD.
    """
    # Lets extract the cartesian orbits and times from the orbits object
    cartesian_orbits = orbits.coordinates.values
    t0 = orbits.coordinates.time.rescale("tdb").mjd()
    t1 = times.rescale("tdb").mjd()
    mu = orbits.coordinates.origin.mu()
    orbit_ids = orbits.orbit_id.to_numpy(zero_copy_only=False)
    object_ids = orbits.object_id.to_numpy(zero_copy_only=False)

    # Lets stack the orbits into a single array shaped by the number of orbits and number of times
    # and then pass this to the vectorized map version of _propagate_2body
    n_orbits = cartesian_orbits.shape[0]
    n_times = len(times)
    orbit_ids_ = np.repeat(orbit_ids, n_times)
    object_ids_ = np.repeat(object_ids, n_times)
    orbits_array_ = np.repeat(cartesian_orbits, n_times, axis=0)
    mu = np.repeat(mu, n_times)
    t0_ = np.repeat(t0, n_times)
    t1_ = np.tile(t1, n_orbits)

    orbits_propagated = _propagate_2body_vmap(
        orbits_array_, t0_, t1_, mu, max_iter, tol
    )
    orbits_propagated = np.array(orbits_propagated)

    if not orbits.coordinates.covariance.is_all_nan():
        cartesian_covariances = orbits.coordinates.covariance.to_matrix()
        covariances_array_ = np.repeat(cartesian_covariances, n_times, axis=0)

        cartesian_covariances = transform_covariances_jacobian(
            orbits_array_,
            covariances_array_,
            _propagate_2body,
            in_axes=(0, 0, 0, 0, None, None),
            out_axes=0,
            t0=t0_,
            t1=t1_,
            mu=mu,
            max_iter=max_iter,
            tol=tol,
        )
        cartesian_covariances = CoordinateCovariances.from_matrix(cartesian_covariances)

    else:
        cartesian_covariances = None

    origin_code = np.empty(n_orbits * n_times, dtype="object")
    origin_code.fill("SUN")

    orbits_propagated = Orbits.from_kwargs(
        orbit_id=orbit_ids_,
        object_id=object_ids_,
        coordinates=CartesianCoordinates.from_kwargs(
            x=orbits_propagated[:, 0],
            y=orbits_propagated[:, 1],
            z=orbits_propagated[:, 2],
            vx=orbits_propagated[:, 3],
            vy=orbits_propagated[:, 4],
            vz=orbits_propagated[:, 5],
            covariance=cartesian_covariances,
            time=Timestamp.from_mjd(t1_, scale="tdb"),
            origin=Origin.from_kwargs(code=origin_code),
            frame="ecliptic",
        ),
    )

    return orbits_propagated



import logging
from typing import List, OrderedDict

import numpy as np
import numpy.typing as npt
from astroquery.jplsbdb import SBDB

from ...coordinates.cometary import CometaryCoordinates
from ...coordinates.covariances import CoordinateCovariances, sigmas_to_covariances
from ...coordinates.origin import Origin
from ...time import Timestamp
from ..orbits import Orbits

logger = logging.getLogger(__name__)


def _convert_SBDB_covariances(
    sbdb_covariances: npt.ArrayLike,
) -> npt.ArrayLike:
    """
    Convert SBDB covariance matrices to Cometary covariance matrices.

    Parameters
    ----------
    sbdb_covariances : `~numpy.ndarray` (N, 6, 6)
        Covariance matrices pulled from JPL's Small Body Database Browser.

    Returns
    -------
    covariances : `~numpy.ndarray` (N, 6, 6)
        Cometary covariance matrices.
    """
    covariances = np.zeros_like(sbdb_covariances)
    # sigma_q{x}
    covariances[:, 0, 0] = sbdb_covariances[:, 1, 1]  # sigma_qq
    covariances[:, 1, 0] = covariances[:, 0, 1] = sbdb_covariances[:, 0, 1]  # sigma_qe
    covariances[:, 2, 0] = covariances[:, 0, 2] = sbdb_covariances[:, 5, 1]  # sigma_qi
    covariances[:, 3, 0] = covariances[:, 0, 3] = sbdb_covariances[
        :, 3, 1
    ]  # sigma_qraan
    covariances[:, 4, 0] = covariances[:, 0, 4] = sbdb_covariances[:, 4, 1]  # sigma_qap
    covariances[:, 5, 0] = covariances[:, 0, 5] = sbdb_covariances[:, 2, 1]  # sigma_qtp

    # sigma_e{x}
    covariances[:, 1, 1] = sbdb_covariances[:, 0, 0]  # sigma_ee
    covariances[:, 2, 1] = covariances[:, 1, 2] = sbdb_covariances[:, 5, 0]  # sigma_ei
    covariances[:, 3, 1] = covariances[:, 1, 3] = sbdb_covariances[
        :, 3, 0
    ]  # sigma_eraan
    covariances[:, 4, 1] = covariances[:, 1, 4] = sbdb_covariances[:, 4, 0]  # sigma_eap
    covariances[:, 5, 1] = covariances[:, 1, 5] = sbdb_covariances[:, 2, 0]  # sigma_etp

    # sigma_i{x}
    covariances[:, 2, 2] = sbdb_covariances[:, 5, 5]  # sigma_ii
    covariances[:, 3, 2] = covariances[:, 2, 3] = sbdb_covariances[
        :, 3, 5
    ]  # sigma_iraan
    covariances[:, 4, 2] = covariances[:, 2, 4] = sbdb_covariances[:, 4, 5]  # sigma_iap
    covariances[:, 5, 2] = covariances[:, 2, 5] = sbdb_covariances[:, 2, 5]  # sigma_itp

    # sigma_raan{x}
    covariances[:, 3, 3] = sbdb_covariances[:, 3, 3]  # sigma_raanraan
    covariances[:, 4, 3] = covariances[:, 3, 4] = sbdb_covariances[
        :, 4, 3
    ]  # sigma_raanap
    covariances[:, 5, 3] = covariances[:, 3, 5] = sbdb_covariances[
        :, 2, 3
    ]  # sigma_raantp

    # sigma_ap{x}
    covariances[:, 4, 4] = sbdb_covariances[:, 4, 4]  # sigma_apap
    covariances[:, 5, 4] = covariances[:, 4, 5] = sbdb_covariances[
        :, 2, 4
    ]  # sigma_aptp

    # sigma_tp{x}
    covariances[:, 5, 5] = sbdb_covariances[:, 2, 2]  # sigma_tptp

    return covariances


def _get_sbdb_elements(obj_ids: List[str]) -> List[OrderedDict]:
    """
    Get orbital elements and other object properties
    from JPL's Small Body Database Browser.

    Parameters
    ----------
    obj_ids : List
        Object IDs to query.

    Returns
    -------
    results : List
        List of dictionaries containing orbital elements and other object properties.
    """
    results = []
    for obj_id in obj_ids:
        result = SBDB.query(
            obj_id,
            covariance="mat",
            id_type="search",
            full_precision=True,
            solution_epoch=False,
        )
        results.append(result)

    return results


def query_sbdb(ids: npt.ArrayLike) -> Orbits:
    """
    Query JPL's Small-Body Database (SBDB) for orbits. The epoch at
    which the orbits are returned are near the epoch as published by the
    Minor Planet Center.

    By default, the orbit's covariance matrices are also queried for. If they
    are not available, then the 1-sigma uncertainties are used to construct
    the covariance matrices.

    Parameters
    ----------
    ids : list
        List of object IDs to query.

    Returns
    -------
    orbits : `~adam_core.orbits.orbits.Orbits`
        Orbits object containing the queried orbits.

    Raises
    ------
    NotFoundError: If any of the queries object IDs are not found.
    """
    results = _get_sbdb_elements(ids)

    orbit_ids = []
    object_ids = []
    classes = []
    coords_cometary = np.zeros((len(results), 6), dtype=np.float64)
    covariances_sbdb = np.zeros((len(results), 6, 6), dtype=np.float64)
    times = np.zeros((len(results)), dtype=np.float64)

    for i, result in enumerate(results):
        if "object" not in result:
            raise NotFoundError("object {} was not found", ids[i])

        orbit_ids.append(f"{i:05d}")
        object_ids.append(result["object"]["fullname"])
        classes.append(result["object"]["orbit_class"]["code"])

        orbit = result["orbit"]
        elements = orbit["elements"]
        epoch = orbit["epoch"]
        if "covariance" in orbit:
            labels = orbit["covariance"]["labels"]
            if len(labels) > 6:
                logger.debug(
                    "Covariance matrix has more parameters than just orbital elements. "
                    "Ignoring non-orbital elements in covariance matrix."
                )
                labels = labels[:6]

            expected_labels = ["e", "q", "tp", "node", "peri", "i"]
            if labels != expected_labels:
                raise ValueError(
                    "Expected covariance matrix labels to be {expected_labels} instead got {labels}."
                )

            # Limit covariances to just the orbital elements
            # The SBDB API documentation states that physical parameter covariances
            # are appended to the rows and columns of the covariance matrix with the
            # orbital elements remaining in the first 6 rows and columns.
            # See: Orbit Subsection: covariance in https://ssd-api.jpl.nasa.gov/doc/sbdb.html
            covariances_sbdb[i, :, :] = orbit["covariance"]["data"][:6, :6]

            if "elements" in orbit["covariance"]:
                # If elements is provided inside covariance, then it's
                # the elements at the epoch which was used to
                # calculate covariance, so we should prefer it.
                elements = orbit["covariance"]["elements"]
                epoch = orbit["covariance"]["epoch"]

        else:
            sigmas = np.array(
                [
                    [
                        elements["e_sig"],
                        elements["q_sig"].value,
                        elements["tp_sig"].value,
                        elements["om_sig"].value,
                        elements["w_sig"].value,
                        elements["i_sig"].value,
                    ]
                ]
            )
            covariances_sbdb[i, :, :] = sigmas_to_covariances(sigmas).filled()[0]

        times[i] = epoch.value
        coords_cometary[i, 0] = elements["q"].value
        coords_cometary[i, 1] = elements["e"]
        coords_cometary[i, 2] = elements["i"].value
        coords_cometary[i, 3] = elements["om"].value
        coords_cometary[i, 4] = elements["w"].value
        coords_cometary[i, 5] = (
            Timestamp.from_jd([elements["tp"].value], scale="tdb").mjd()[0].as_py()
        )

    covariances_cometary = _convert_SBDB_covariances(covariances_sbdb)
    times = Timestamp.from_jd(times, scale="tdb")
    origin = Origin.from_kwargs(code=["SUN" for i in range(len(times))])
    frame = "ecliptic"
    coordinates = CometaryCoordinates.from_kwargs(
        time=times,
        q=coords_cometary[:, 0],
        e=coords_cometary[:, 1],
        i=coords_cometary[:, 2],
        raan=coords_cometary[:, 3],
        ap=coords_cometary[:, 4],
        tp=coords_cometary[:, 5],
        covariance=CoordinateCovariances.from_matrix(covariances_cometary),
        origin=origin,
        frame=frame,
    )

    orbit_ids = np.array(orbit_ids, dtype="object")
    object_ids = np.array(object_ids, dtype="object")
    classes = np.array(classes)

    return Orbits.from_kwargs(
        orbit_id=orbit_ids, object_id=object_ids, coordinates=coordinates.to_cartesian()
    )


class NotFoundError(Exception):
    def __init__(self, message, object_id):
        self.message = message
        self.object_id = object_id

    def __str__(self):
        return self.message



import numpy as np
import pytest
from astropy.time import Time

from ..mpc import (
    convert_mpc_packed_dates,
    pack_mpc_designation,
    pack_numbered_designation,
    pack_provisional_designation,
    pack_survey_designation,
    unpack_mpc_designation,
    unpack_numbered_designation,
    unpack_provisional_designation,
    unpack_survey_designation,
)

# --- Tests last updated: 22-03-2023
# Packed : Unpacked
# Taken from: https://minorplanetcenter.net/iau/info/PackedDes.html
# Provisional Minor Planet Designations
PROVISIONAL_DESIGNATIONS_UP2P = {
    "1995 XA": "J95X00A",
    "1995 XL1": "J95X01L",
    "1995 FB13": "J95F13B",
    "1998 SQ108": "J98SA8Q",
    "1998 SV127": "J98SC7V",
    "1998 SS162": "J98SG2S",
    "2099 AZ193": "K99AJ3Z",
    "2008 AA360": "K08Aa0A",
    "2007 TA418": "K07Tf8A",
}
# Provisional Minor Planet Designations (Surveys)
SURVEY_DESIGNATIONS_UP2P = {
    "2040 P-L": "PLS2040",
    "3138 T-1": "T1S3138",
    "1010 T-2": "T2S1010",
    "4101 T-3": "T3S4101",
}
# Permanent Minor Planet Designations
NUMBERED_DESIGNATION_UP2P = {
    "3202": "03202",
    "50000": "50000",
    "100345": "A0345",
    "360017": "a0017",
    "203289": "K3289",
    "620000": "~0000",
    "620061": "~000z",
    "3140113": "~AZaz",
    "15396335": "~zzzz",
}
DESIGNATIONS_UP2P = {
    **PROVISIONAL_DESIGNATIONS_UP2P,
    **SURVEY_DESIGNATIONS_UP2P,
    **NUMBERED_DESIGNATION_UP2P,
}


def test_convert_mpc_packed_dates():
    # Use a few modified examples from https://minorplanetcenter.net/iau/info/PackedDates.html
    # and test conversion from packed form to MJDs
    isot_tt = np.array(
        [
            "1996-01-01",
            "1996-01-10",
            "1996-09-30",
            "1996-10-01",
            "2001-10-22",
            "2001-10-22T00:00:00.0000",
            "2001-10-22T12:00:00.0000",
            "1996-09-30T18:00:00.0000",
            "1996-09-30T18:45:00.0000",
        ]
    )

    pf_tt = np.array(
        [
            "J9611",
            "J961A",
            "J969U",
            "J96A1",
            "K01AM",
            "K01AM",
            "K01AM5",
            "J969U75",
            "J969U78125",
        ]
    )

    mjd_actual = convert_mpc_packed_dates(pf_tt)
    mjd_desired = Time(isot_tt, format="isot", scale="tt")

    np.testing.assert_equal(mjd_actual.tt.mjd, mjd_desired.tt.mjd)
    return


def test_unpack_numbered_designation():
    # Test unpacking of numbered designations
    for designation, designation_pf in NUMBERED_DESIGNATION_UP2P.items():
        assert unpack_numbered_designation(designation_pf) == designation


def test_unpack_numbered_designation_raises():
    # Test invalid unpacking of provisional designations as numbered designations
    for designation_pf in PROVISIONAL_DESIGNATIONS_UP2P:
        with pytest.raises(ValueError):
            unpack_numbered_designation(designation_pf)

    # Test invalid unpacking of survey designations as numbered designations
    for designation_pf in SURVEY_DESIGNATIONS_UP2P:
        with pytest.raises(ValueError):
            unpack_numbered_designation(designation_pf)


def test_unpack_provisional_designation():
    # Test unpacking of provisional designations
    for designation, designation_pf in PROVISIONAL_DESIGNATIONS_UP2P.items():
        assert unpack_provisional_designation(designation_pf) == designation


def test_unpack_provisional_designation_raises():
    # Test invalid unpacking of numbered designations as provisional designations
    for designation_pf in NUMBERED_DESIGNATION_UP2P:
        with pytest.raises(ValueError):
            unpack_provisional_designation(designation_pf)

    # Test invalid unpacking of survey designations as provisional designations
    for designation_pf in SURVEY_DESIGNATIONS_UP2P:
        with pytest.raises(ValueError):
            unpack_provisional_designation(designation_pf)


def test_unpack_survey_designation():
    # Test unpacking of survey designations
    for designation, designation_pf in SURVEY_DESIGNATIONS_UP2P.items():
        assert unpack_survey_designation(designation_pf) == designation


def test_unpack_survey_designation_raises():
    # Test invalid unpacking of numbered designations as survey designations
    for designation_pf in NUMBERED_DESIGNATION_UP2P:
        with pytest.raises(ValueError):
            unpack_survey_designation(designation_pf)

    # Test invalid unpacking of provisional designations as survey designations
    for designation_pf in PROVISIONAL_DESIGNATIONS_UP2P:
        with pytest.raises(ValueError):
            unpack_survey_designation(designation_pf)


def test_pack_numbered_designation():
    # Test packing of numbered designations
    for designation, designation_pf in NUMBERED_DESIGNATION_UP2P.items():
        assert pack_numbered_designation(designation) == designation_pf


def test_pack_numbered_designation_raises():
    # Test invalid packing of provisional designations as numbered designations
    for designation in PROVISIONAL_DESIGNATIONS_UP2P:
        with pytest.raises(ValueError):
            pack_numbered_designation(designation)

    # Test invalid packing of survey designations as numbered designations
    for designation in SURVEY_DESIGNATIONS_UP2P:
        with pytest.raises(ValueError):
            pack_numbered_designation(designation)


def test_pack_provisional_designation():
    # Test packing of provisional designations
    for designation, designation_pf in PROVISIONAL_DESIGNATIONS_UP2P.items():
        assert pack_provisional_designation(designation) == designation_pf


def test_pack_provisional_designation_raises():
    # Test invalid packing of numbered designations as provisional designations
    for designation in NUMBERED_DESIGNATION_UP2P:
        with pytest.raises(ValueError):
            pack_provisional_designation(designation)

    # Test invalid packing of survey designations as provisional designations
    for designation in SURVEY_DESIGNATIONS_UP2P:
        with pytest.raises(ValueError):
            pack_provisional_designation(designation)


def test_pack_survey_designation():
    # Test packing of survey designations
    for designation, designation_pf in SURVEY_DESIGNATIONS_UP2P.items():
        assert pack_survey_designation(designation) == designation_pf


def test_pack_survey_designation_raises():
    # Test invalid packing of numbered designations as survey designations
    for designation in NUMBERED_DESIGNATION_UP2P:
        with pytest.raises(ValueError):
            pack_survey_designation(designation)

    # Test invalid packing of provisional designations as survey designations
    for designation in PROVISIONAL_DESIGNATIONS_UP2P:
        with pytest.raises(ValueError):
            pack_survey_designation(designation)


def test_unpack_mpc_designation():
    # Test unpacking of packed form designations
    for designation_pf, designation in DESIGNATIONS_UP2P.items():
        assert unpack_mpc_designation(designation) == designation_pf


def test_pack_mpc_designation():
    # Test packing of unpacked designations
    for designation, designation_pf in DESIGNATIONS_UP2P.items():
        assert pack_mpc_designation(designation) == designation_pf



from typing import Tuple

import jax.numpy as jnp
import numpy as np
from jax import jit, lax, vmap

from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.covariances import (
    CoordinateCovariances,
    transform_covariances_jacobian,
)
from ..coordinates.origin import Origin, OriginCodes
from ..coordinates.spherical import SphericalCoordinates
from ..coordinates.transform import _cartesian_to_spherical, transform_coordinates
from ..observers.observers import Observers
from ..orbits.ephemeris import Ephemeris
from ..orbits.orbits import Orbits
from .aberrations import _add_light_time, add_stellar_aberration


@jit
def _generate_ephemeris_2body(
    propagated_orbit: np.ndarray,
    observation_time: float,
    observer_coordinates: jnp.ndarray,
    mu: float,
    lt_tol: float = 1e-10,
    max_iter: int = 100,
    tol: float = 1e-15,
    stellar_aberration: bool = False,
) -> Tuple[jnp.ndarray, jnp.float64]:
    """
    Given a propagated orbit, generate its on-sky ephemeris as viewed from the observer.
    This function calculates the light time delay between the propagated orbit and the observer,
    and then propagates the orbit backward by that amount to when the light from object was actually
    emitted towards the observer ("astrometric coordinates").

    The motion of the observer in an inertial frame will cause an object
    to appear in a different location than its true location, this is known as
    stellar aberration (often referred to in combination with other aberrations as "apparent
    coordinates"). Stellar aberration can optionally be applied after
    light time correction has been added but it should not be necessary when comparing to ephemerides
    of solar system small bodies extracted from astrometric catalogs. The stars to which the
    catalog is calibrated undergo the same aberration as the moving objects as seen from the observer.

    If stellar aberration is applied then the velocity of the input orbits are unmodified, only the position
    vector is modified with stellar aberration.

    For more details on aberrations see:
        https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/FORTRAN/req/abcorr.html
        https://ssd.jpl.nasa.gov/horizons/manual.html#defs

    Parameters
    ----------
    propagated_orbit : `~jax.numpy.ndarray` (6)
        Barycentric Cartesian orbit propagated to the given time.
    observation_time : float
        Epoch at which orbit and observer coordinates are defined.
    observer_coordinates : `~jax.numpy.ndarray` (3)
        Barycentric Cartesian observer coordinates.
    mu : float (1)
        Gravitational parameter (GM) of the attracting body in units of
        AU**3 / d**2.
    lt_tol : float, optional
        Calculate aberration to within this value in time (units of days).
    max_iter : int, optional
        Maximum number of iterations over which to converge for propagation.
    tol : float, optional
        Numerical tolerance to which to compute universal anomaly during propagation using the Newtown-Raphson
        method.
    stellar_aberration : bool, optional
        Apply stellar aberration to the ephemerides.

    Returns
    -------
    ephemeris_spherical : `~jax.numpy.ndarray` (6)
        Topocentric Spherical ephemeris.
    lt : float
        Light time correction (t0 - corrected_t0).
    """
    # Add light time correction
    propagated_orbits_aberrated, light_time = _add_light_time(
        propagated_orbit,
        observation_time,
        observer_coordinates[0:3],
        lt_tol=lt_tol,
        mu=mu,
        max_iter=max_iter,
        tol=tol,
    )

    # Calculate topocentric coordinates
    topocentric_coordinates = propagated_orbits_aberrated - observer_coordinates

    # Apply stellar aberration to topocentric coordinates
    topocentric_coordinates = lax.cond(
        stellar_aberration,
        lambda topocentric_coords: topocentric_coords.at[0:3].set(
            add_stellar_aberration(
                propagated_orbits_aberrated.reshape(1, -1),
                observer_coordinates.reshape(1, -1),
            )[0],
        ),
        lambda topocentric_coords: topocentric_coords,
        topocentric_coordinates,
    )

    # Convert to spherical coordinates
    ephemeris_spherical = _cartesian_to_spherical(topocentric_coordinates)

    return ephemeris_spherical, light_time


# Vectorization Map: _generate_ephemeris_2body
_generate_ephemeris_2body_vmap = jit(
    vmap(
        _generate_ephemeris_2body,
        in_axes=(0, 0, 0, 0, None, None, None, None),
        out_axes=(0, 0),
    )
)


def generate_ephemeris_2body(
    propagated_orbits: Orbits,
    observers: Observers,
    lt_tol: float = 1e-10,
    max_iter: int = 1000,
    tol: float = 1e-15,
    stellar_aberration: bool = False,
) -> Ephemeris:
    """
    Generate on-sky ephemerides for each propagated orbit as viewed by the observers.
    This function calculates the light time delay between the propagated orbit and the observer,
    and then propagates the orbit backward by that amount to when the light from object was actually
    emitted towards the observer ("astrometric coordinates").

    The motion of the observer in an inertial frame will cause an object
    to appear in a different location than its true location, this is known as
    stellar aberration (often referred to in combination with other aberrations as "apparent
    coordinates"). Stellar aberration can optionally be applied after
    light time correction has been added but it should not be necessary when comparing to ephemerides
    of solar system small bodies extracted from astrometric catalogs. The stars to which the
    catalog is calibrated undergo the same aberration as the moving objects as seen from the observer.

    If stellar aberration is applied then the velocity of the input orbits are unmodified, only the position
    vector is modified with stellar aberration.

    For more details on aberrations see:
        https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/FORTRAN/req/abcorr.html
        https://ssd.jpl.nasa.gov/horizons/manual.html#defs


    Parameters
    ----------
    propagated_orbits : `~adam_core.orbits.orbits.Orbits` (N)
        Propagated orbits.
    observers : `~adam_core.observers.observers.Observers` (N)
        Observers for which to generate ephemerides. Orbits should already have been
        propagated to the same times as the observers.
    lt_tol : float, optional
        Calculate aberration to within this value in time (units of days).
    mu : float, optional
        Gravitational parameter (GM) of the attracting body in units of
        AU**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge for propagation.
    tol : float, optional
        Numerical tolerance to which to compute universal anomaly during propagation using the Newtown-Raphson
        method.
    stellar_aberration : bool, optional
        Apply stellar aberration to the ephemerides.

    Returns
    -------
    ephemeris : `~adam_core.orbits.ephemeris.Ephemeris` (N)
        Topocentric ephemerides for each propagated orbit as observed by the given observers.
    """
    # Transform both the orbits and observers to the barycenter if they are not already.
    propagated_orbits_barycentric = propagated_orbits.set_column(
        "coordinates",
        transform_coordinates(
            propagated_orbits.coordinates,
            CartesianCoordinates,
            frame_out="ecliptic",
            origin_out=OriginCodes.SOLAR_SYSTEM_BARYCENTER,
        ),
    )
    observers_barycentric = observers.set_column(
        "coordinates",
        transform_coordinates(
            observers.coordinates,
            CartesianCoordinates,
            frame_out="ecliptic",
            origin_out=OriginCodes.SOLAR_SYSTEM_BARYCENTER,
        ),
    )

    # Stack the observer coordinates and codes for each orbit in the propagated orbits
    num_orbits = len(propagated_orbits_barycentric.orbit_id.unique())
    observer_coordinates = np.tile(
        observers_barycentric.coordinates.values, (num_orbits, 1)
    )
    observer_codes = np.tile(observers.code.to_numpy(zero_copy_only=False), num_orbits)
    mu = observers_barycentric.coordinates.origin.mu()
    mu = np.tile(mu, num_orbits)

    times = propagated_orbits.coordinates.time.to_astropy()
    ephemeris_spherical, light_time = _generate_ephemeris_2body_vmap(
        propagated_orbits_barycentric.coordinates.values,
        times.mjd,
        observer_coordinates,
        mu,
        lt_tol,
        max_iter,
        tol,
        stellar_aberration,
    )
    ephemeris_spherical = np.array(ephemeris_spherical)
    light_time = np.array(light_time)

    if not propagated_orbits.coordinates.covariance.is_all_nan():

        cartesian_covariances = propagated_orbits.coordinates.covariance.to_matrix()
        covariances_spherical = transform_covariances_jacobian(
            propagated_orbits.coordinates.values,
            cartesian_covariances,
            _generate_ephemeris_2body,
            in_axes=(0, 0, 0, 0, None, None, None, None),
            out_axes=(0, 0),
            observation_times=times.utc.mjd,
            observer_coordinates=observer_coordinates,
            mu=mu,
            lt_tol=lt_tol,
            max_iter=max_iter,
            tol=tol,
            stellar_aberration=stellar_aberration,
        )
        covariances_spherical = CoordinateCovariances.from_matrix(
            np.array(covariances_spherical)
        )

    else:
        covariances_spherical = None

    spherical_coordinates = SphericalCoordinates.from_kwargs(
        time=propagated_orbits.coordinates.time,
        rho=ephemeris_spherical[:, 0],
        lon=ephemeris_spherical[:, 1],
        lat=ephemeris_spherical[:, 2],
        vrho=ephemeris_spherical[:, 3],
        vlon=ephemeris_spherical[:, 4],
        vlat=ephemeris_spherical[:, 5],
        covariance=covariances_spherical,
        origin=Origin.from_kwargs(code=observer_codes),
        frame="ecliptic",
    )

    # Rotate the spherical coordinates from the ecliptic frame
    # to the equatorial frame
    spherical_coordinates = transform_coordinates(
        spherical_coordinates, SphericalCoordinates, frame_out="equatorial"
    )

    return Ephemeris.from_kwargs(
        orbit_id=propagated_orbits_barycentric.orbit_id,
        object_id=propagated_orbits_barycentric.object_id,
        coordinates=spherical_coordinates,
        light_time=light_time,
    )



from importlib.resources import files
from typing import List

import numpy as np
import pandas as pd
import pyarrow as pa
import quivr as qv

from adam_core.observers import Observers
from adam_core.orbits import Orbits
from adam_core.orbits.query import query_horizons, query_horizons_ephemeris, query_sbdb
from adam_core.orbits.query.horizons import (
    _get_horizons_elements,
    _get_horizons_vectors,
)
from adam_core.time import Timestamp


def _get_orbital_elements(
    object_ids: List[str], time: Timestamp, location: str, refplane: str = "ecliptic"
) -> pd.DataFrame:
    """
    Get orbital elements as Cartesian, Cometary, and Keplerian representations from JPL Horizons.

    Parameters
    ----------
    object_id : List[str]
        Object IDs to
    epoch : Timestamp
        Epoch at which to query orbital elements.
    location : str
        Location of the observer (in this case typically "@sun" or "@ssb"
        for barycentric or heliocentric elements, respectively).

    Returns
    -------
    horizons_elements : `~pandas.DataFrame`
        DataFrame containing orbital elements in different representations for each object ID.
    """
    vectors_df = _get_horizons_vectors(
        object_ids,
        time,
        location=location,
        id_type="smallbody",
        aberrations="geometric",
        refplane=refplane,
    )
    vectors_df = vectors_df[
        ["targetname", "datetime_jd", "x", "y", "z", "vx", "vy", "vz"]
    ]

    elements_df = _get_horizons_elements(
        object_ids,
        time,
        location=location,
        id_type="smallbody",
        refplane=refplane,
    )
    elements_df = elements_df[
        [
            "targetname",
            "datetime_jd",
            "a",
            "q",
            "Q",
            "e",
            "incl",
            "Omega",
            "w",
            "M",
            "nu",
            "n",
            "Tp_jd",
            "P",
        ]
    ]

    horizons_elements = vectors_df.merge(elements_df, on=["targetname", "datetime_jd"])

    horizons_elements.insert(
        1, "mjd_tdb", Timestamp.from_jd(horizons_elements["datetime_jd"]).mjd()
    )
    horizons_elements.insert(
        len(horizons_elements.columns),
        "tp_mjd",
        Timestamp.from_jd(horizons_elements["Tp_jd"]).mjd(),
    )
    horizons_elements.drop(columns=["datetime_jd", "Tp_jd"], inplace=True)
    return horizons_elements


if __name__ == "__main__":
    # Read sample object IDs from the included objects.csv file
    objects_file = files("adam_core.utils.helpers.data").joinpath("objects.csv")
    objects_df = pd.read_csv(objects_file)

    # Query for the orbital elements from SBDB and save to a file
    object_ids = objects_df["object_id"].values

    orbits = query_sbdb(object_ids)

    # Rename 'Oumuamua to match the name in the Horizons database (1I/'Oumuamua (A/2017 U1))
    object_ids = orbits.object_id.to_numpy(zero_copy_only=False)
    orbits = orbits.set_column(
        "object_id",
        np.where(
            object_ids == "'Oumuamua (A/2017 U1)",
            "1I/'Oumuamua (A/2017 U1)",
            object_ids,
        ),
    )

    orbits.to_parquet(files("adam_core.utils.helpers.data").joinpath("orbits.parquet"))

    # Query for the orbital elements in different representations from JPL Horizons and save to a file
    for frame in ["ecliptic", "equatorial"]:
        if frame == "equatorial":
            refplane = "earth"
            frame_out = "eq"
        else:
            refplane = "ecliptic"
            frame_out = "ec"

        for origin in ["@sun", "@ssb"]:

            horizons_elements_dfs = []
            for orbit in orbits:
                prov_designation = (
                    orbit.object_id.to_numpy(zero_copy_only=False)[0]
                    .split("(")
                    .pop()
                    .split(")")[0]
                )

                horizons_elements_df = _get_orbital_elements(
                    [prov_designation],
                    orbit.coordinates.time,
                    location=origin,
                    refplane=refplane,
                )
                horizons_elements_dfs.append(horizons_elements_df)

            horizons_elements_df = pd.concat(horizons_elements_dfs)

            if origin == "@sun":
                horizons_elements_df.to_csv(
                    files("adam_core.utils.helpers.data").joinpath(
                        f"elements_sun_{frame_out}.csv"
                    ),
                    index=False,
                )
            else:
                horizons_elements_df.to_csv(
                    files("adam_core.utils.helpers.data").joinpath(
                        f"elements_ssb_{frame_out}.csv"
                    ),
                    index=False,
                )

    # Lets query for propagated Horizons state vectors
    propagated_orbits_list = []
    ephemeris_dfs = []

    # Create an array of delta times to propagate the orbits and
    # get ephemerides relative to the epoch at which the orbits are defined
    total_days = 60
    half_arc = total_days / 2
    # Make it so there are 3 observations every 2 days
    dts = np.arange(-half_arc, half_arc, 2)
    dts = np.concatenate([dts, dts + 1 / 48, dts + 1 / 24])
    dts.sort()
    num_dts = len(dts)

    for i, orbit in enumerate(orbits):
        # Get the provisional designation
        prov_designation = (
            orbit.object_id.to_numpy(zero_copy_only=False)[0]
            .split("(")
            .pop()
            .split(")")[0]
        )

        # Extract times from orbit and propagate +/- 30 days
        times = orbit.coordinates.time.add_fractional_days(dts)

        # Query for propagated Horizons state vectors
        propagated_orbit = query_horizons([prov_designation], times)
        propagated_orbit = propagated_orbit.set_column(
            "orbit_id", pa.array([f"{i:05d}" for _ in range(len(times))])
        )

        # Define two observers one at the Rubin Observatory and one at CTIO
        observer_X05 = Observers.from_code("X05", times[: num_dts // 2])
        observer_W84 = Observers.from_code("W84", times[num_dts // 2 :])
        observers = qv.concatenate([observer_X05, observer_W84])

        # Query for ephemeris at the same times
        ephemeris = query_horizons_ephemeris([prov_designation], observers)
        ephemeris["orbit_id"] = [f"{i:05d}" for _ in range(len(times))]

        propagated_orbits_list.append(propagated_orbit)
        ephemeris_dfs.append(ephemeris)

    propagated_orbits: Orbits = qv.concatenate(propagated_orbits_list)
    propagated_orbits.to_parquet(
        files("adam_core.utils.helpers.data").joinpath("propagated_orbits.parquet"),
    )

    ephemeris_df = pd.concat(ephemeris_dfs, ignore_index=True)
    ephemeris_df.to_csv(
        files("adam_core.utils.helpers.data").joinpath("ephemeris.csv"), index=False
    )



import logging
from abc import ABC, abstractmethod
from typing import List, Literal, Optional, Type, Union

import numpy as np
import numpy.typing as npt
import quivr as qv

from adam_core.ray_cluster import initialize_use_ray

from ..observers.observers import Observers
from ..orbits.ephemeris import Ephemeris
from ..orbits.orbits import Orbits
from ..orbits.variants import VariantEphemeris, VariantOrbits
from ..time import Timestamp
from .utils import _iterate_chunks

logger = logging.getLogger(__name__)

RAY_INSTALLED = False
try:
    import ray
    from ray import ObjectRef

    RAY_INSTALLED = True
except ImportError:
    pass

TimestampType = Union[Timestamp, ObjectRef]
OrbitType = Union[Orbits, VariantOrbits, ObjectRef]
EphemerisType = Union[Ephemeris, VariantOrbits, ObjectRef]
ObserverType = Union[Observers, ObjectRef]


def propagation_worker(
    orbits: Union[Orbits, VariantOrbits],
    times: Timestamp,
    propagator: Type["Propagator"],
    **kwargs,
) -> Union[Orbits, VariantOrbits]:
    prop = propagator(**kwargs)
    propagated = prop._propagate_orbits(orbits, times)
    return propagated


def ephemeris_worker(
    orbits: Union[Orbits, VariantOrbits],
    observers: Observers,
    propagator: Type["Propagator"],
    **kwargs,
) -> Union[Ephemeris, VariantOrbits]:
    prop = propagator(**kwargs)
    ephemeris = prop._generate_ephemeris(orbits, observers)
    return ephemeris


if RAY_INSTALLED:

    @ray.remote
    def propagation_worker_ray(
        idx: npt.NDArray[np.int64],
        orbits: OrbitType,
        times: OrbitType,
        propagator: Type["Propagator"],
        **kwargs,
    ) -> OrbitType:
        prop = propagator(**kwargs)
        orbits_chunk = orbits.take(idx)
        propagated = prop._propagate_orbits(orbits_chunk, times)
        return propagated

    @ray.remote
    def ephemeris_worker_ray(
        idx: npt.NDArray[np.int64],
        orbits: OrbitType,
        observers: ObserverType,
        propagator: Type["Propagator"],
        **kwargs,
    ) -> EphemerisType:
        prop = propagator(**kwargs)
        orbits_chunk = orbits.take(idx)
        ephemeris = prop._generate_ephemeris(orbits_chunk, observers)
        return ephemeris


class EphemerisMixin:
    """
    Mixin with signature for generating ephemerides.
    Subclasses should implement the _generate_ephemeris method.
    """

    @abstractmethod
    def _generate_ephemeris(
        self, orbits: EphemerisType, observers: ObserverType
    ) -> EphemerisType:
        """
        Generate ephemerides for the given orbits as observed by
        the observers.

        THIS FUNCTION SHOULD BE DEFINED BY THE USER.
        """
        pass

    def generate_ephemeris(
        self,
        orbits: OrbitType,
        observers: ObserverType,
        covariance: bool = False,
        covariance_method: Literal[
            "auto", "sigma-point", "monte-carlo"
        ] = "monte-carlo",
        num_samples: int = 1000,
        chunk_size: int = 100,
        max_processes: Optional[int] = 1,
    ) -> Ephemeris:
        """
        Generate ephemerides for each orbit in orbits as observed by each observer
        in observers.

        Parameters
        ----------
        orbits : `~adam_core.orbits.orbits.Orbits` (N)
            Orbits for which to generate ephemerides.
        observers : `~adam_core.observers.observers.Observers` (M)
            Observers for which to generate the ephemerides of each
            orbit.
        covariance: bool, optional
            Propagate the covariance matrices of the orbits. This is done by sampling the
            orbits from their covariance matrices and propagating each sample and for each
            sample also generating ephemerides. The covariance
            of the ephemerides is then the covariance of the samples.
        covariance_method : {'sigma-point', 'monte-carlo', 'auto'}, optional
            The method to use for sampling the covariance matrix. If 'auto' is selected then the method
            will be automatically selected based on the covariance matrix. The default is 'monte-carlo'.
        num_samples : int, optional
            The number of samples to draw when sampling with monte-carlo.
        chunk_size : int, optional
            Number of orbits to send to each job.
        max_processes : int or None, optional
            Number of processes to launch. If None then the number of
            processes will be equal to the number of cores on the machine. If 1
            then no multiprocessing will be used. If "ray" is the parallel_backend and a ray instance
            is initialized already then this argument is ignored.
        parallel_backend : {'cf', 'ray'}, optional
            The parallel backend to use. 'cf' uses concurrent.futures and 'ray' uses ray. The default is 'cf'.
            To use ray, ray must be installed.

        Returns
        -------
        ephemeris : `~adam_core.orbits.ephemeris.Ephemeris` (M)
            Predicted ephemerides for each orbit observed by each
            observer.
        """
        # Check if we need to propagate orbit variants so we can propagate covariance
        # matrices

        if max_processes is None or max_processes > 1:
            ephemeris_list: List[Ephemeris] = []
            variants_list: List[VariantEphemeris] = []

            if RAY_INSTALLED is False:
                raise ImportError(
                    "Ray must be installed to use the ray parallel backend"
                )

            initialize_use_ray(num_cpus=max_processes)

            # Add orbits and observers to object store if
            # they haven't already been added
            if not isinstance(observers, ObjectRef):
                observers_ref = ray.put(observers)
            else:
                observers_ref = observers

            if not isinstance(orbits, ObjectRef):
                orbits_ref = ray.put(orbits)
            else:
                orbits_ref = orbits
                # We need to dereference the orbits ObjectRef so we can
                # check its length for chunking and determine
                # if we need to propagate variants
                orbits = ray.get(orbits_ref)

            # Create futures
            futures = []
            idx = np.arange(0, len(orbits))
            for idx_chunk in _iterate_chunks(idx, chunk_size):
                futures.append(
                    ephemeris_worker_ray.remote(
                        idx_chunk,
                        orbits_ref,
                        observers_ref,
                        self.__class__,
                        **self.__dict__,
                    )
                )

            # Add variants to futures (if we have any)
            if covariance is True and not orbits.coordinates.covariance.is_all_nan():
                variants = VariantOrbits.create(
                    orbits, method=covariance_method, num_samples=num_samples
                )

                # Add variants to object store
                variants_ref = ray.put(variants)

                idx = np.arange(0, len(variants))
                for variant_chunk in _iterate_chunks(idx, chunk_size):
                    idx_chunk = ray.put(variant_chunk)
                    futures.append(
                        ephemeris_worker_ray.remote(
                            idx_chunk,
                            variants_ref,
                            observers_ref,
                            self.__class__,
                            **self.__dict__,
                        )
                    )

            # Get results as they finish (we sort later)
            unfinished = futures
            while unfinished:
                finished, unfinished = ray.wait(unfinished, num_returns=1)
                result = ray.get(finished[0])
                if isinstance(result, Ephemeris):
                    ephemeris_list.append(result)
                elif isinstance(result, VariantEphemeris):
                    variants_list.append(result)
                else:
                    raise ValueError(
                        f"Unexpected result type from ephemeris worker: {type(result)}"
                    )

            ephemeris = qv.concatenate(ephemeris_list)
            if len(variants_list) > 0:
                ephemeris_variants = qv.concatenate(variants_list)
            else:
                ephemeris_variants = None

        else:
            ephemeris = self._generate_ephemeris(orbits, observers)

            if covariance is True and not orbits.coordinates.covariance.is_all_nan():
                variants = VariantOrbits.create(
                    orbits, method=covariance_method, num_samples=num_samples
                )
                ephemeris_variants = self._generate_ephemeris(variants, observers)
            else:
                ephemeris_variants = None

        if ephemeris_variants is not None:
            ephemeris = ephemeris_variants.collapse(ephemeris)

        return ephemeris.sort_by(
            [
                "orbit_id",
                "coordinates.time.days",
                "coordinates.time.nanos",
                "coordinates.origin.code",
            ]
        )


class Propagator(ABC):
    """
    Abstract class for propagating orbits and related functions.

    Subclasses should implement the _propagate_orbits.
    For additional functions, subclasses can add abstract mixins.

    Important: subclasses should be pickleable! As this class uses multiprocessing
    to parallelize propagation and ephemeris generation. This means that
    subclasses should not have any unpickleable attributes.
    """

    @abstractmethod
    def _propagate_orbits(self, orbits: OrbitType, times: TimestampType) -> OrbitType:
        """
        Propagate orbits to times.

        THIS FUNCTION SHOULD BE DEFINED BY THE USER.
        """
        pass

    def propagate_orbits(
        self,
        orbits: OrbitType,
        times: TimestampType,
        covariance: bool = False,
        covariance_method: Literal[
            "auto", "sigma-point", "monte-carlo"
        ] = "monte-carlo",
        num_samples: int = 1000,
        chunk_size: int = 100,
        max_processes: Optional[int] = 1,
    ) -> Orbits:
        """
        Propagate each orbit in orbits to each time in times.

        This method handles parallelization of the propagation of the orbits.
        Subclasses may override this method to modify parallelization behavior.

        Parameters
        ----------
        orbits : `~adam_core.orbits.orbits.Orbits` (N)
            Orbits to propagate.
        times : Timestamp (M)
            Times to which to propagate orbits.
        covariance : bool, optional
            Propagate the covariance matrices of the orbits. This is done by sampling the
            orbits from their covariance matrices and propagating each sample. The covariance
            of the propagated orbits is then the covariance of the samples.
        covariance_method : {'sigma-point', 'monte-carlo', 'auto'}, optional
            The method to use for sampling the covariance matrix. If 'auto' is selected then the method
            will be automatically selected based on the covariance matrix. The default is 'monte-carlo'.
        num_samples : int, optional
            The number of samples to draw when sampling with monte-carlo.
        chunk_size : int, optional
            Number of orbits to send to each job.
        max_processes : int or None, optional
            Maximum number of processes to launch. If None then the number of
            processes will be equal to the number of cores on the machine. If 1
            then no multiprocessing will be used. If "ray" is the parallel_backend and a ray instance
            is initialized already then this argument is ignored.

        Returns
        -------
        propagated : `~adam_core.orbits.orbits.Orbits`
            Propagated orbits.
        """
        if max_processes is None or max_processes > 1:
            propagated_list: List[Orbits] = []
            variants_list: List[VariantOrbits] = []

            if RAY_INSTALLED is False:
                raise ImportError(
                    "Ray must be installed to use the ray parallel backend"
                )

            initialize_use_ray(num_cpus=max_processes)

            # Add orbits and times to object store if
            # they haven't already been added
            if not isinstance(times, ObjectRef):
                times_ref = ray.put(times)
            else:
                times_ref = times

            if not isinstance(orbits, ObjectRef):
                orbits_ref = ray.put(orbits)
            else:
                orbits_ref = orbits
                # We need to dereference the orbits ObjectRef so we can
                # check its length for chunking and determine
                # if we need to propagate variants
                orbits = ray.get(orbits_ref)

            # Create futures
            futures = []
            idx = np.arange(0, len(orbits))
            for idx_chunk in _iterate_chunks(idx, chunk_size):
                futures.append(
                    propagation_worker_ray.remote(
                        idx_chunk,
                        orbits_ref,
                        times_ref,
                        self.__class__,
                        **self.__dict__,
                    )
                )

            # Add variants to propagate to futures
            if covariance is True and not orbits.coordinates.covariance.is_all_nan():
                variants = VariantOrbits.create(
                    orbits, method=covariance_method, num_samples=num_samples
                )
                variants_ref = ray.put(variants)

                idx = np.arange(0, len(variants))
                for variant_chunk in _iterate_chunks(idx, chunk_size):
                    idx_chunk = ray.put(variant_chunk)
                    futures.append(
                        propagation_worker_ray.remote(
                            idx_chunk,
                            variants_ref,
                            times_ref,
                            self.__class__,
                            **self.__dict__,
                        )
                    )

            # Get results as they finish (we sort later)
            unfinished = futures
            while unfinished:
                finished, unfinished = ray.wait(unfinished, num_returns=1)
                result = ray.get(finished[0])
                if isinstance(result, Orbits):
                    propagated_list.append(result)
                elif isinstance(result, VariantOrbits):
                    variants_list.append(result)
                else:
                    raise ValueError(
                        f"Unexpected result type from propagation worker: {type(result)}"
                    )

            # Concatenate propagated orbits
            propagated = qv.concatenate(propagated_list)
            if len(variants_list) > 0:
                propagated_variants = qv.concatenate(variants_list)
            else:
                propagated_variants = None

        else:
            propagated = self._propagate_orbits(orbits, times)

            if covariance is True and not orbits.coordinates.covariance.is_all_nan():
                variants = VariantOrbits.create(
                    orbits, method=covariance_method, num_samples=num_samples
                )
                propagated_variants = self._propagate_orbits(variants, times)
            else:
                propagated_variants = None

        if propagated_variants is not None:
            propagated = propagated_variants.collapse(propagated)

        return propagated.sort_by(
            ["orbit_id", "coordinates.time.days", "coordinates.time.nanos"]
        )



import pytest

from ..orbits import make_real_orbits, make_simple_orbits


def test_make_real_orbits():
    # Test that all sample orbits are read from the CSV file
    orbits = make_real_orbits()
    assert len(orbits) == 28

    # Test that the number of orbits returned is correct
    orbits = make_real_orbits(num_orbits=10)
    assert len(orbits) == 10


def test_make_simple_orbits_raises():
    # Test that a ValueError is raised if num_orbits > 29
    with pytest.raises(ValueError):
        make_real_orbits(num_orbits=29)


def test_make_simple_orbits():
    # Test that the number of orbits returned is correct
    orbits = make_simple_orbits(num_orbits=10)
    assert len(orbits) == 10



import numpy as np

from ..origin import Origin
from ..spherical import SphericalCoordinates


def test_SphericalCoordinates_to_unit_sphere():
    # Test that the to_unit_sphere method works as expected. This method
    # should return a copy of the coordinates with rho set to 1.0 and
    # vrho set to 0.0. If only_missing is True, then only coordinates
    # that have NaN values for rho will be set to 1.0 and coordinates
    # that have NaN values for vrho will be set to 0.0.
    coords = SphericalCoordinates.from_kwargs(
        rho=[1.0, np.NaN, 3.0],
        lon=[30.0, 45.0, 60.0],
        lat=[-60, 0.0, 60.0],
        vrho=[np.NaN, 0.002, -0.01],
        vlon=[0.005, 0.006, 0.007],
        vlat=[-0.0005, 0.0006, -0.007],
        origin=Origin.from_kwargs(code=["SUN", "SUN", "SUN"]),
        frame="ecliptic",
    )

    unit_sphere = coords.to_unit_sphere(only_missing=True)
    np.testing.assert_allclose(unit_sphere.rho.to_numpy(), [1.0, 1.0, 3.0])
    np.testing.assert_allclose(unit_sphere.vrho.to_numpy(), [0.0, 0.002, -0.01])
    np.testing.assert_allclose(unit_sphere.lon.to_numpy(), [30.0, 45.0, 60.0])
    np.testing.assert_allclose(unit_sphere.lat.to_numpy(), [-60.0, 0.0, 60.0])
    np.testing.assert_allclose(unit_sphere.vlon.to_numpy(), [0.005, 0.006, 0.007])
    np.testing.assert_allclose(unit_sphere.vlat.to_numpy(), [-0.0005, 0.0006, -0.007])

    unit_sphere = coords.to_unit_sphere(only_missing=False)
    np.testing.assert_allclose(unit_sphere.rho.to_numpy(), [1.0, 1.0, 1.0])
    np.testing.assert_allclose(unit_sphere.vrho.to_numpy(), [0.0, 0.0, 0.0])
    np.testing.assert_allclose(unit_sphere.lon.to_numpy(), [30.0, 45.0, 60.0])
    np.testing.assert_allclose(unit_sphere.lat.to_numpy(), [-60.0, 0.0, 60.0])
    np.testing.assert_allclose(unit_sphere.vlon.to_numpy(), [0.005, 0.006, 0.007])
    np.testing.assert_allclose(unit_sphere.vlat.to_numpy(), [-0.0005, 0.0006, -0.007])



import numpy as np
from astropy import units as u
from quivr.concat import concatenate

from ...time import Timestamp
from ..cartesian import CartesianCoordinates
from ..origin import Origin, OriginCodes
from ..transform import cartesian_to_origin, transform_coordinates


def assert_coords_equal(
    have: CartesianCoordinates,
    want: CartesianCoordinates,
):
    diff = want.values - have.values

    # Calculate offset in position in mm
    r_diff = np.linalg.norm(diff[:, :3], axis=1) * u.au.to(u.mm)
    # Calculate offset in velocity in nm/s
    v_diff = np.linalg.norm(diff[:, 3:], axis=1) * (u.au / u.d).to(u.nm / u.s)

    # Assert positions are to within 10 mm
    np.testing.assert_array_less(r_diff, 10)
    # Assert velocities are to within 10 nm/s
    np.testing.assert_array_less(v_diff, 10)


def test_cartesian_to_origin(orbital_elements, orbital_elements_barycentric):
    # Test cartesian_to_origin correctly converts between heliocentric and
    # barycentric cartesian coordinates

    cartesian_coordinates_heliocentric = CartesianCoordinates.from_kwargs(
        x=orbital_elements["x"].values,
        y=orbital_elements["y"].values,
        z=orbital_elements["z"].values,
        vx=orbital_elements["vx"].values,
        vy=orbital_elements["vy"].values,
        vz=orbital_elements["vz"].values,
        time=Timestamp.from_mjd(
            orbital_elements["mjd_tdb"].values,
            scale="tdb",
        ),
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN" for i in range(len(orbital_elements))]),
    )

    cartesian_coordinates_barycentric = CartesianCoordinates.from_kwargs(
        x=orbital_elements_barycentric["x"].values,
        y=orbital_elements_barycentric["y"].values,
        z=orbital_elements_barycentric["z"].values,
        vx=orbital_elements_barycentric["vx"].values,
        vy=orbital_elements_barycentric["vy"].values,
        vz=orbital_elements_barycentric["vz"].values,
        time=Timestamp.from_mjd(
            orbital_elements_barycentric["mjd_tdb"].values,
            scale="tdb",
        ),
        frame="ecliptic",
        origin=Origin.from_kwargs(
            code=[
                "SOLAR_SYSTEM_BARYCENTER"
                for i in range(len(orbital_elements_barycentric))
            ]
        ),
    )

    # Convert heliocentric cartesian coordinates to barycentric cartesian coordinates
    cartesian_coordinates_barycentric_actual = cartesian_to_origin(
        cartesian_coordinates_heliocentric,
        OriginCodes.SOLAR_SYSTEM_BARYCENTER,
    )

    assert_coords_equal(
        cartesian_coordinates_barycentric, cartesian_coordinates_barycentric
    )

    # Convert barycentric cartesian coordinates to heliocentric cartesian coordinates
    cartesian_coordinates_heliocentric_actual = cartesian_to_origin(
        cartesian_coordinates_barycentric,
        OriginCodes.SUN,
    )

    assert_coords_equal(
        cartesian_coordinates_heliocentric_actual, cartesian_coordinates_heliocentric
    )

    # Now lets complicate matters by concatenating heliocentric and barycentric
    # cartesian coordinates and converting them to barycentric cartesian coordinates
    cartesian_coordinates_mixed = concatenate(
        [cartesian_coordinates_heliocentric, cartesian_coordinates_barycentric]
    )

    cartesian_coordinates_barycentric_actual = cartesian_to_origin(
        cartesian_coordinates_mixed,
        OriginCodes.SOLAR_SYSTEM_BARYCENTER,
    )

    cartesian_coordinates_barycentric_ = concatenate(
        [cartesian_coordinates_barycentric for i in range(2)]
    )

    assert_coords_equal(
        cartesian_coordinates_barycentric_actual, cartesian_coordinates_barycentric_
    )

    # Now lets complicate matters again by concatenating heliocentric and barycentric
    # cartesian coordinates and converting them to heliocentric cartesian coordinates
    cartesian_coordinates_mixed = concatenate(
        [cartesian_coordinates_heliocentric, cartesian_coordinates_barycentric]
    )

    cartesian_coordinates_heliocentric_actual = cartesian_to_origin(
        cartesian_coordinates_mixed,
        OriginCodes.SUN,
    )

    cartesian_coordinates_heliocentric_ = concatenate(
        [cartesian_coordinates_heliocentric for i in range(2)]
    )

    assert_coords_equal(
        cartesian_coordinates_heliocentric_actual, cartesian_coordinates_heliocentric_
    )


def test_transform_coordinates_origin(orbital_elements, orbital_elements_barycentric):
    # Test transform_coordinates correctly converts between heliocentric and
    # barycentric cartesian coordinates

    cartesian_coordinates_heliocentric = CartesianCoordinates.from_kwargs(
        x=orbital_elements["x"].values,
        y=orbital_elements["y"].values,
        z=orbital_elements["z"].values,
        vx=orbital_elements["vx"].values,
        vy=orbital_elements["vy"].values,
        vz=orbital_elements["vz"].values,
        time=Timestamp.from_mjd(
            orbital_elements["mjd_tdb"].values,
            scale="tdb",
        ),
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN" for i in range(len(orbital_elements))]),
    )

    cartesian_coordinates_barycentric = CartesianCoordinates.from_kwargs(
        x=orbital_elements_barycentric["x"].values,
        y=orbital_elements_barycentric["y"].values,
        z=orbital_elements_barycentric["z"].values,
        vx=orbital_elements_barycentric["vx"].values,
        vy=orbital_elements_barycentric["vy"].values,
        vz=orbital_elements_barycentric["vz"].values,
        time=Timestamp.from_mjd(
            orbital_elements_barycentric["mjd_tdb"].values,
            scale="tdb",
        ),
        frame="ecliptic",
        origin=Origin.from_kwargs(
            code=[
                "SOLAR_SYSTEM_BARYCENTER"
                for i in range(len(orbital_elements_barycentric))
            ]
        ),
    )

    # Convert heliocentric cartesian coordinates to barycentric cartesian coordinates
    cartesian_coordinates_barycentric_actual = transform_coordinates(
        cartesian_coordinates_heliocentric,
        origin_out=OriginCodes.SOLAR_SYSTEM_BARYCENTER,
    )

    assert_coords_equal(
        cartesian_coordinates_barycentric, cartesian_coordinates_barycentric
    )

    # Convert barycentric cartesian coordinates to heliocentric cartesian coordinates
    cartesian_coordinates_heliocentric_actual = transform_coordinates(
        cartesian_coordinates_barycentric,
        origin_out=OriginCodes.SUN,
    )

    assert_coords_equal(
        cartesian_coordinates_heliocentric_actual, cartesian_coordinates_heliocentric
    )

    # Now lets complicate matters by concatenating heliocentric and barycentric
    # cartesian coordinates and converting them to barycentric cartesian coordinates
    cartesian_coordinates_mixed = concatenate(
        [cartesian_coordinates_heliocentric, cartesian_coordinates_barycentric]
    )

    cartesian_coordinates_barycentric_actual = transform_coordinates(
        cartesian_coordinates_mixed,
        origin_out=OriginCodes.SOLAR_SYSTEM_BARYCENTER,
    )

    cartesian_coordinates_barycentric_ = concatenate(
        [cartesian_coordinates_barycentric for i in range(2)]
    )

    assert_coords_equal(
        cartesian_coordinates_barycentric_actual, cartesian_coordinates_barycentric_
    )

    # Now lets complicate matters again by concatenating heliocentric and barycentric
    # cartesian coordinates and converting them to heliocentric cartesian coordinates
    cartesian_coordinates_mixed = concatenate(
        [cartesian_coordinates_heliocentric, cartesian_coordinates_barycentric]
    )

    cartesian_coordinates_heliocentric_actual = transform_coordinates(
        cartesian_coordinates_mixed,
        origin_out=OriginCodes.SUN,
    )

    cartesian_coordinates_heliocentric_ = concatenate(
        [cartesian_coordinates_heliocentric for i in range(2)]
    )

    assert_coords_equal(
        cartesian_coordinates_heliocentric_actual, cartesian_coordinates_heliocentric_
    )



from __future__ import annotations

import astropy.time
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import quivr as qv

SCALES = {
    "tai",
    "tt",
    "ut1",
    "utc",
    "tdb",
}

# The Modified Julian Date of the J2000 epoch in TDB scale:
_J2000_TDB_MJD = 51544.5


class Timestamp(qv.Table):
    # Scale, the rate at which time passes:
    scale = qv.StringAttribute(default="tai")

    # Days since MJD epoch (1858-11-17T00:00:00):
    days = qv.Int64Column()

    # Nanos since start of day:
    nanos = qv.Int64Column()

    def micros(self) -> pa.Int64Array:
        return pc.divide(self.nanos, 1_000)

    def millis(self) -> pa.Int64Array:
        return pc.divide(self.nanos, 1_000_000)

    def seconds(self) -> pa.Int64Array:
        return pc.divide(self.nanos, 1_000_000_000)

    def mjd(self) -> pa.lib.DoubleArray:
        return pc.add(self.days, self.fractional_days())

    def jd(self) -> pa.lib.DoubleArray:
        return pc.add(self.mjd(), 2400000.5)

    def et(self) -> pa.lib.DoubleArray:
        """
        Returns the times as ET seconds in a pyarrow array.
        """
        tdb = self.rescale("tdb")

        mjd = tdb.mjd()
        return pc.multiply(pc.subtract(mjd, _J2000_TDB_MJD), 86400)

    def to_numpy(self) -> np.ndarray:
        """
        Returns the times as TDB MJDs in a numpy array.
        """
        return self.rescale("tdb").mjd().to_numpy(False)

    @classmethod
    def from_iso8601(
        cls, iso: pa.lib.StringArray | list[str], scale="utc"
    ) -> Timestamp:
        """
        Create a Timestamp from ISO 8601 strings (for example, '2020-01-02T14:15:16').
        """
        return cls.from_astropy(astropy.time.Time(iso, format="isot", scale=scale))

    @classmethod
    def from_mjd(cls, mjd: pa.lib.DoubleArray, scale: str = "tai") -> Timestamp:
        days = pc.floor(mjd)
        fractional_days = pc.subtract(mjd, days)
        days = pc.cast(days, pa.int64())
        nanos = pc.cast(pc.round(pc.multiply(fractional_days, 86400 * 1e9)), pa.int64())
        return cls.from_kwargs(days=days, nanos=nanos, scale=scale)

    @classmethod
    def from_jd(cls, jd: pa.lib.DoubleArray, scale: str = "tai") -> Timestamp:
        return cls.from_mjd(pc.subtract(jd, 2400000.5), scale)

    def fractional_days(self) -> pa.lib.DoubleArray:
        return pc.divide(self.nanos, 86400 * 1e9)

    def rounded(self, precision: str = "ns") -> Timestamp:
        if precision == "ns":
            return self
        elif precision == "us":
            nanos = pc.multiply(self.micros(), 1_000)
        elif precision == "ms":
            nanos = pc.multiply(self.millis(), 1_000_000)
        elif precision == "s":
            nanos = pc.multiply(self.seconds(), 1_000_000_000)
        else:
            raise ValueError(f"Unsupported precision: {precision}")
        return self.set_column("nanos", nanos)

    def equals(self, other: Timestamp, precision: str = "ns") -> pa.BooleanArray:
        if self.scale != other.scale:
            raise ValueError("Cannot compare timestamps with different scales")
        if len(other) == 1:
            return self.equals_scalar(other.days[0], other.nanos[0], precision)
        else:
            return self.equals_array(other, precision)

    def equals_scalar(
        self, days: int, nanos: int, precision: str = "ns"
    ) -> pa.BooleanArray:
        delta_days, delta_nanos = self.difference_scalar(days, nanos)
        if precision == "ns":
            max_deviation = 0
        elif precision == "us":
            max_deviation = 999
        elif precision == "ms":
            max_deviation = 999_999
        elif precision == "s":
            max_deviation = 999_999_999
        else:
            raise ValueError(f"Unsupported precision: {precision}")
        return _duration_arrays_within_tolerance(delta_days, delta_nanos, max_deviation)

    def equals_array(self, other: Timestamp, precision: str = "ns") -> pa.BooleanArray:
        """
        Compare two Timestamps, returning a BooleanArray indicating
        whether each element is equal.

        The Timestamps must have the same scale, and the same length.
        """
        if self.scale != other.scale:
            raise ValueError("Cannot compare timestamps with different scales")
        if len(self) != len(other):
            raise ValueError("Timestamps must have the same length")

        delta_days, delta_nanos = self.difference(other)
        if precision == "ns":
            max_deviation = 0
        elif precision == "us":
            max_deviation = 999
        elif precision == "ms":
            max_deviation = 999_999
        elif precision == "s":
            max_deviation = 999_999_999
        else:
            raise ValueError(f"Unsupported precision: {precision}")
        return _duration_arrays_within_tolerance(delta_days, delta_nanos, max_deviation)

    def max(self) -> Timestamp:
        """
        Compute the maximum time.

        Returns
        -------
        max_time : `~adam_core.time.Timestamp`
            The maximum time. If there are multiple maximum times,
            one of them is returned.
        """
        # Compute the maximum day
        max_day = pc.max(self.days)
        days_mask = pc.equal(self.days, max_day)

        # Compute the maximum nanos for the maximum day
        max_nanos = pc.max(self.nanos.filter(days_mask))
        nanos_mask = pc.equal(self.nanos, max_nanos)

        return self.apply_mask(pc.and_(days_mask, nanos_mask))[0]

    def min(self) -> Timestamp:
        """
        Compute the minimum time.

        Returns
        -------
        min_time : `~adam_core.time.Timestamp`
            The minimum time. If there are multiple minimum times,
            one of them is returned.
        """
        # Compute the minimum day
        min_day = pc.min(self.days)
        days_mask = pc.equal(self.days, min_day)

        # Compute the minimum nanos for the minimum day
        min_nanos = pc.min(self.nanos.filter(days_mask))
        nanos_mask = pc.equal(self.nanos, min_nanos)

        return self.apply_mask(pc.and_(days_mask, nanos_mask))[0]

    @classmethod
    def from_astropy(cls, astropy_time: astropy.time.Time) -> Timestamp:
        """Convert an astropy time to a quivr timestamp.

        This is a lossy conversion, since astropy uses floating point
        to represent times, while quivr uses integers.

        The astropy time must use a scale supported by quivr. The
        supported scales are "tai", "tt", "ut1", "utc", and "tdb".

        """
        if astropy_time.scale not in SCALES:
            raise ValueError(f"Unsupported scale: {astropy_time.scale}")
        if astropy_time.isscalar:
            return cls._from_astropy_scalar(astropy_time)

        jd1, jd2 = astropy_time.jd1, astropy_time.jd2
        days, remainder = divmod(jd1 - 2400000.5, 1)
        remainder += jd2

        mask = remainder < 0
        np.add(remainder, 1, where=mask, out=remainder)
        np.subtract(days, 1, where=mask, out=days)

        mask = remainder >= 1
        np.subtract(remainder, 1, where=mask, out=remainder)
        np.add(days, 1, where=mask, out=days)

        nanos = np.round(remainder * 86400 * 1e9)
        return cls.from_kwargs(
            scale=astropy_time.scale,
            days=days.astype(np.int64),
            nanos=nanos,
        )

    @classmethod
    def _from_astropy_scalar(cls, astropy_time: astropy.time.Time) -> Timestamp:
        """Astropy times can be scalar-valued, which requires
        separate handling because the numpy array functions won't
        work.
        """
        jd1, jd2 = astropy_time.jd1, astropy_time.jd2
        days, remainder = divmod(jd1 - 2400000.5, 1)
        remainder += jd2

        if remainder < 0:
            remainder += 1
            days -= 1

        if remainder >= 1:
            remainder -= 1
            days += 1

        nanos = round(remainder * 86400 * 1e9)
        return cls.from_kwargs(
            scale=astropy_time.scale,
            days=[int(days)],
            nanos=[nanos],
        )

    def to_astropy(self) -> astropy.time.Time:
        """
        Convert the timestamp to an astropy time.
        """
        fractional_days = self.fractional_days()
        return astropy.time.Time(
            val=self.days,
            val2=fractional_days,
            format="mjd",
            scale=self.scale,
        )

    def add_nanos(
        self, nanos: pa.lib.Int64Array | int, check_range: bool = True
    ) -> Timestamp:
        """
        Add nanoseconds to the timestamp. Negative nanoseconds are
        allowed.

        Parameters
        ----------
        nanos : The nanoseconds to add. Can be a scalar or an array of
            the same length as the timestamp. Must be in the range [-86400e9, 86400e9).
        check_range : If True, check that the nanoseconds are in the
            range [-86400e9, 86400e9). If False, the caller is
            responsible for ensuring that the nanoseconds are in the
            correct range.
        """
        if check_range:
            if isinstance(nanos, int):
                if not -86400e9 <= nanos < 86400e9:
                    raise ValueError("Nanoseconds out of range")
            else:
                if not pc.all(
                    pc.and_(pc.greater_equal(nanos, -86400e9), pc.less(nanos, 86400e9))
                ).as_py():
                    raise ValueError("Nanoseconds out of range")

        nanos = pc.add_checked(self.nanos, nanos)
        overflows = pc.greater_equal(nanos, 86400 * 1e9)
        underflows = pc.less(nanos, 0)

        mask = pa.StructArray.from_arrays(
            [overflows, underflows], names=["overflows", "underflows"]
        )
        nanos = pc.case_when(
            mask,
            pc.subtract(nanos, int(86400 * 1e9)),
            pc.add(nanos, int(86400 * 1e9)),
            nanos,
        )

        days = pc.case_when(
            mask,
            pc.add(self.days, 1),
            pc.subtract(self.days, 1),
            self.days,
        )
        v1 = self.set_column("days", days)
        v2 = v1.set_column("nanos", nanos)
        return v2

    def add_seconds(
        self, seconds: pa.lib.Int64Array | int | pa.DoubleArray | float
    ) -> Timestamp:
        """
        Add seconds to the timestamp. Negative seconds are supported.

        Parameters
        ----------
        seconds : The seconds to add. Can be a scalar or an array of
            the same length as the timestamp. Must be in the range [-86400, 86400).

        See Also
        --------
        add_nanos : Add nanoseconds to the timestamp. This method includes
            a 'check_range' parameter that allows the caller to disable range
            checking for performance reasons.
        """
        nanos = pc.cast(pc.round(pc.multiply(seconds, 1_000_000_000)), pa.int64())
        return self.add_nanos(nanos)

    def add_millis(self, millis: pa.lib.Int64Array | int) -> Timestamp:
        """
        Add milliseconds to the timestamp. Negative milliseconds are
        supported.

        Parameters
        ----------
        millis : The milliseconds to add. Can be a scalar or an array of
            the same length as the timestamp. Must be in the range [-86400e3, 86400e3).

        See Also
        --------
        add_nanos : Add nanoseconds to the timestamp. This method includes
            a 'check_range' parameter that allows the caller to disable range
            checking for performance reasons.
        """
        nanos = pc.cast(pc.round(pc.multiply(millis, 1_000_000)), pa.int64())
        return self.add_nanos(nanos)

    def add_micros(self, micros: pa.lib.Int64Array | int) -> Timestamp:
        """
        Add microseconds to the timestamp. Negative microseconds are
        supported.

        Parameters
        ----------
        micros : The microseconds to add. Can be a scalar or an array of
            the same length as the timestamp. Must be in the range [-86400e6, 86400e6).

        See Also
        --------
        add_nanos : Add nanoseconds to the timestamp. This method includes
            a 'check_range' parameter that allows the caller to disable range
            checking for performance reasons.
        """
        nanos = pc.cast(pc.round(pc.multiply(micros, 1_000)), pa.int64())
        return self.add_nanos(nanos)

    def add_days(self, days: pa.lib.Int64Array | int) -> Timestamp:
        """Add days to the timestamp.

        Parameters
        ----------
        days : The days to add. Can be a scalar or an array of the
            same length as the timestamp. Use negative values to
            subtract days.

        """
        return self.set_column("days", pc.add(self.days, days))

    def add_fractional_days(
        self, fractional_days: pa.lib.DoubleArray | float
    ) -> Timestamp:
        """
        Add fractional days to the timestamp.

        Parameters
        ----------
        fractional_days : The fractional days to add. Can be a scalar
            or an array of the same length as the timestamp. Use
            negative values to subtract fractional days.
        """
        day_part = pc.floor(fractional_days)
        nano_part = pc.subtract(fractional_days, day_part)

        days = pc.cast(day_part, pa.int64())
        nanos = pc.cast(pc.multiply(nano_part, 86400 * 1e9), pa.int64())

        return self.add_days(days).add_nanos(nanos)

    def difference_scalar(
        self, days: int, nanos: int
    ) -> tuple[pa.Int64Array, pa.Int64Array]:
        """
        Compute the difference between this timestamp and a scalar
        timestamp.

        The difference is computed as (self - scalar). The result is
        presented as a tuple of (days, nanos). The nanos value is
        always non-negative, in the range [0, 86400e9).

        Parameters
        ----------
        days : The days of the scalar timestamp.
        nanos : The nanoseconds of the scalar timestamp.

        Returns
        -------
        days : The difference in days. This value can be negative.
        nanos : The difference in nanoseconds. This value is always
            non-negative, in the range [0, 86400e9).

        Examples
        --------
        >>> from adam_core.time import Timestamp
        >>> ts = Timestamp.from_kwargs(days=[0, 1, 2], nanos=[200, 0, 100])
        >>> have_days, have_nanos = ts.difference_scalar(1, 100)
        >>> have_days.to_pylist()
        [-1, -1, 1]
        >>> have_nanos.to_pylist()
        [100, 86399999999900, 0]

        """
        days1 = pc.subtract(self.days, days)
        nanos1 = pc.subtract(self.nanos, nanos)
        overflows = pc.greater_equal(nanos1, 86400 * 1e9)
        underflows = pc.less(nanos1, 0)
        mask = pa.StructArray.from_arrays(
            [overflows, underflows], names=["overflows", "underflows"]
        )
        nanos2 = pc.case_when(
            mask,
            pc.subtract(nanos1, int(86400 * 1e9)),
            pc.add(nanos1, int(86400 * 1e9)),
            nanos1,
        )
        days2 = pc.case_when(
            mask,
            pc.add(days1, 1),
            pc.subtract(days1, 1),
            days1,
        )
        return days2, nanos2

    def difference(self, other: Timestamp) -> tuple[pa.Int64Array, pa.Int64Array]:
        """
        Compute the element-wise difference between this timestamp and another.
        """
        if self.scale != other.scale:
            raise ValueError(
                "Cannot compute difference between timestamps with different scales"
            )
        days1 = pc.subtract(self.days, other.days)
        nanos1 = pc.subtract(self.nanos, other.nanos)

        overflows = pc.greater_equal(nanos1, 86400 * 1e9)
        underflows = pc.less(nanos1, 0)
        mask = pa.StructArray.from_arrays(
            [overflows, underflows], names=["overflows", "underflows"]
        )
        nanos2 = pc.case_when(
            mask,
            pc.subtract(nanos1, int(86400 * 1e9)),
            pc.add(nanos1, int(86400 * 1e9)),
            nanos1,
        )
        days2 = pc.case_when(
            mask,
            pc.add(days1, 1),
            pc.subtract(days1, 1),
            days1,
        )
        return days2, nanos2

    def unique(self) -> Timestamp:
        """Return a new Timestamp table containing only the unique
        elements from self. Order is not necessarily preserved.

        """
        uniqued = self.table.group_by(["days", "nanos"]).aggregate([])
        uniqued = uniqued.replace_schema_metadata(self.table.schema.metadata)
        return Timestamp.from_pyarrow(uniqued)

    def rescale(self, new_scale: str) -> Timestamp:
        if self.scale == new_scale:
            return self
        elif new_scale == "tai":
            return Timestamp.from_astropy(self.to_astropy().tai)
        elif new_scale == "utc":
            return Timestamp.from_astropy(self.to_astropy().utc)
        elif new_scale == "tt":
            return Timestamp.from_astropy(self.to_astropy().tt)
        elif new_scale == "ut1":
            return Timestamp.from_astropy(self.to_astropy().ut1)
        elif new_scale == "tdb":
            return Timestamp.from_astropy(self.to_astropy().tdb)
        else:
            raise ValueError("Unknown scale: {}".format(new_scale))

    def link(
        self, other: Timestamp, precision: str = "ns"
    ) -> qv.MultiKeyLinkage[Timestamp, Timestamp]:
        """
        Link this Timestamp to another. The default precision is nanoseconds, but if
        other precisions are desired then both this class and the other Timestamp will
        be rounded to the desired precision.

        If the timescales are different, the other Timestamp will be rescaled to
        this Timestamp's timescale.

        Parameters
        ----------
        other : The Timestamp to link to.
        precision : The precision to use when linking. The default is 'ns'.

        Returns
        -------
        linkage : A MultiKeyLinkage object that can be used to join the two Timestamps.
        """
        if self.scale != other.scale:
            other = other.rescale(self.scale)

        rounded = self.rounded(precision)
        other_rounded = other.rounded(precision)

        left_keys = {"days": rounded.days, "nanos": rounded.nanos}
        right_keys = {"days": other_rounded.days, "nanos": other_rounded.nanos}
        return qv.MultiKeyLinkage(self, other, left_keys, right_keys)


def _duration_arrays_within_tolerance(
    delta_days: pa.Int64Array, delta_nanos: pa.Int64Array, max_nanos_deviation: int
) -> pa.BooleanArray:
    """Return a boolean array indicating whether the delta_days and delta_nanos
    arrays are within the specified tolerance.

    The max_nanos_deviation should be the maximum number of
    nanoseconds that the the two arrays can deviate to still be
    considered 'within tolerance'.
    """
    if max_nanos_deviation == 0:
        return pc.and_(pc.equal(delta_days, 0), pc.equal(delta_nanos, 0))

    cond1 = pc.and_(
        pc.equal(delta_days, 0), pc.less(pc.abs(delta_nanos), max_nanos_deviation)
    )
    cond2 = pc.and_(
        pc.equal(delta_days, -1),
        pc.greater_equal(pc.abs(delta_nanos), 86400 * 1e9 - max_nanos_deviation),
    )
    return pc.or_(cond1, cond2)



import numpy as np
import pytest

from ...utils.helpers.orbits import make_real_orbits
from ..covariances import (
    CoordinateCovariances,
    sample_covariance_random,
    sample_covariance_sigma_points,
    weighted_covariance,
    weighted_mean,
)


def test_sample_covariance_sigma_points():
    # Get a sample of real orbits and test that sigma point sampling
    # allows the state vector and its covariance to be reconstructed
    orbits = make_real_orbits()

    # Limit to first 10 orbits
    for orbit in orbits[:10]:
        mean = orbit.coordinates.values[0]
        covariance = orbit.coordinates.covariance.to_matrix()[0]

        samples, W, W_cov = sample_covariance_sigma_points(mean, covariance)

        # In a 6 dimensional space we expect 13 sigma point samples
        assert len(samples) == 13
        np.testing.assert_almost_equal(np.sum(W), 1.0)
        np.testing.assert_almost_equal(np.sum(W_cov), 1.0)
        # The first sample should be the mean
        np.testing.assert_equal(samples[0], mean)
        # The first weight should be 0.0
        assert W[0] == 0.0
        # The first weight for the covariance should be 0
        # since beta = 0 internally
        assert W_cov[0] == 0.0

        # Reconstruct the mean and covariance and test that they match
        # the original inputs to within 1e-14
        mean_sg = weighted_mean(samples, W)
        covariance_sg = weighted_covariance(mean, samples, W_cov)

        np.testing.assert_allclose(mean_sg, mean, rtol=0, atol=1e-14)
        np.testing.assert_allclose(covariance_sg, covariance, rtol=0, atol=1e-14)


def test_sample_covariance_random():
    # Get a sample of real orbits and test that random sampling
    # allows the state vector and its covariance to be reconstructed
    orbits = make_real_orbits()

    np.random.seed(0)
    # Limit to first 10 orbits
    for orbit in orbits[:10]:
        mean = orbit.coordinates.values[0]
        covariance = orbit.coordinates.covariance.to_matrix()[0]

        samples, W, W_cov = sample_covariance_random(mean, covariance, 1000000)

        # In a 6 dimensional space we expect 1000000 samples
        assert len(samples) == 1000000
        np.testing.assert_almost_equal(np.sum(W), 1.0)
        np.testing.assert_almost_equal(np.sum(W_cov), 1.0)

        # Reconstruct the mean and covariance and test that they match
        # the original inputs to within 1e-8 and 1e-14 respectively
        mean_rs = weighted_mean(samples, W)
        covariance_rs = weighted_covariance(mean, samples, W_cov)

        # Note how many samples are needed to get the covariance to
        # match to within 1e-14... (the same tolerance as sigma point sampling)
        np.testing.assert_allclose(mean_rs, mean, rtol=0, atol=1e-8)
        np.testing.assert_allclose(covariance_rs, covariance, rtol=0, atol=1e-14)


def test_CoordinateCovariances_from_sigmas():
    # Given an array of sigmas, test that the
    # covariance matrices are correctly calculated
    sigmas = np.array(
        [
            [2.0, 3.0, 4.0, 5.0, 6.0, 7.0],
            [3.0, 4.0, 5.0, 6.0, 7.0, 8.0],
        ]
    )
    covariances = np.array(
        [
            [
                [4.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 9.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 16.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 25.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 36.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 49.0],
            ],
            [
                [9.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 16.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 25.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 36.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 49.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 64.0],
            ],
        ]
    )

    cov = CoordinateCovariances.from_sigmas(sigmas)
    np.testing.assert_equal(cov.to_matrix(), covariances)


def test_CoordinateCovariances_to_from_matrix():
    # Given an array of covariances test that the
    # covariance matrices are correctly returned
    covariances = np.zeros((10, 6, 6))
    for i in range(10):
        covariances[i, :, :] = np.diag(np.arange(1, 7))

    cov = CoordinateCovariances.from_matrix(covariances)
    np.testing.assert_equal(cov.to_matrix(), covariances)

    # Test when covariances are mixed with None and np.array
    covariances = [None, np.ones((6, 6)).flatten()]
    cov = CoordinateCovariances.from_kwargs(values=covariances)
    cov_expected = np.ones((2, 6, 6))
    cov_expected[0, :, :] = np.NaN
    np.testing.assert_equal(cov.to_matrix(), cov_expected)

    # Test when covariances are only None
    covariances = [None, None]
    cov = CoordinateCovariances.from_kwargs(values=covariances)
    cov_expected = np.full((2, 6, 6), np.NaN)
    np.testing.assert_equal(cov.to_matrix(), cov_expected)


def test_CoordinateCovariances_is_all_nan():
    # Test that all_nan convenience method works as intended
    # No NaNs at all
    covariances = np.ones((10, 6, 6))
    cov = CoordinateCovariances.from_matrix(covariances)
    assert not cov.is_all_nan()

    # Single NaN value
    covariances = np.ones((10, 6, 6))
    covariances[0, 0, 0] = np.nan
    cov = CoordinateCovariances.from_matrix(covariances)
    assert not cov.is_all_nan()

    # Null covariance (None)
    covariances = [None, np.ones((6, 6)).flatten()]
    cov = CoordinateCovariances.from_kwargs(values=covariances)
    assert not cov.is_all_nan()

    # All NaNs
    covariances = np.ones((10, 6, 6)) * np.nan
    cov = CoordinateCovariances.from_matrix(covariances)
    assert cov.is_all_nan()


def test_CoordinateCovariances_from_matrix_invalid_shape():
    # Test that an error is raised when the shape of the input
    # covariances is invalid
    covariances = np.zeros((10, 6, 5))
    with pytest.raises(ValueError):
        CoordinateCovariances.from_matrix(covariances)

    covariances = np.zeros((6, 5))
    with pytest.raises(ValueError):
        CoordinateCovariances.from_matrix(covariances)



import jax.numpy as jnp
from jax import config, jit

config.update("jax_enable_x64", True)


@jit
def solve_barker(M: float) -> float:
    """
    Solve Barker's equation for true anomaly given parabolic mean
    anomaly.

    Parameters
    ----------
    M : float
        Parabolic mean anomaly (equal to sqrt(mu / (2 q^3))(t0 - tp)).

    Returns
    -------
    nu : float
        True anomaly in radians.

    References
    ----------
    [1] Curtis, H. D. (2014). Orbital Mechanics for Engineering Students. 3rd ed.,
        Elsevier Ltd. ISBN-13: 978-0080977478
    """
    # Equation 3.32 in Curtis (2014) [1]
    nu = 2 * jnp.arctan(
        (3 * M + jnp.sqrt((3 * M) ** 2 + 1)) ** (1 / 3)
        - (3 * M + jnp.sqrt((3 * M) ** 2 + 1)) ** (-1 / 3)
    )
    return nu



import numpy as np
import numpy.testing as npt

from ..sbdb import _convert_SBDB_covariances


def test_convert_SBDB_covariances():
    sbdb_format = np.array(
        [
            ["e_e", "e_q", "e_tp", "e_raan", "e_ap", "e_i"],
            ["q_e", "q_q", "q_tp", "q_raan", "q_ap", "q_i"],
            ["tp_e", "tp_q", "tp_tp", "tp_raan", "tp_ap", "tp_i"],
            ["raan_e", "raan_q", "raan_tp", "raan_raan", "raan_ap", "raan_i"],
            ["ap_e", "ap_q", "ap_tp", "ap_raan", "ap_ap", "ap_i"],
            ["i_e", "i_q", "i_tp", "i_raan", "i_ap", "i_i"],
        ]
    )
    adam_format = np.array(
        [
            ["q_q", "q_e", "q_i", "q_raan", "q_ap", "q_tp"],
            ["e_q", "e_e", "e_i", "e_raan", "e_ap", "e_tp"],
            ["i_q", "i_e", "i_i", "i_raan", "i_ap", "i_tp"],
            ["raan_q", "raan_e", "raan_i", "raan_raan", "raan_ap", "raan_tp"],
            ["ap_q", "ap_e", "ap_i", "ap_raan", "ap_ap", "ap_tp"],
            ["tp_q", "tp_e", "tp_i", "tp_raan", "tp_ap", "tp_tp"],
        ]
    )
    sbdb_format = np.array([sbdb_format])
    adam_format = np.array([adam_format])

    cometary_covariances = _convert_SBDB_covariances(sbdb_format)
    # Some of the symmetric terms will be in reverse format: for example
    # q_e will be e_q where it should be q_e. So for places where
    # the order is reversed, let's try to flip them around and then lets
    # check for equality with the expected ADAM format.

    flip_mask = np.where(cometary_covariances != adam_format)
    for i, j, k in zip(*flip_mask):
        cometary_covariances[i, j, k] = "_".join(
            cometary_covariances[i, j, k].split("_")[::-1]
        )

    npt.assert_equal(cometary_covariances, adam_format)



import os

import numpy as np
import pyarrow as pa
import pytest
import quivr as qv

try:
    from adam_core.propagator.adam_pyoorb import PYOORBPropagator
except ImportError:
    PYOORBPropagator = None

from ..evaluate import evaluate_orbits


@pytest.mark.skipif(
    os.environ.get("OORB_DATA") is None, reason="OORB_DATA environment variable not set"
)
@pytest.mark.skipif(PYOORBPropagator is None, reason="PYOORBPropagator not available")
def test_evaluate_orbits(pure_iod_orbit):
    # Test that evaluate_orbit correctly calculates residuals and other
    # parameters for an input orbit
    orbit, orbit_members, observations = pure_iod_orbit
    propagator = PYOORBPropagator()

    # Concatenate the orbit three times to test we can handle multiple orbits
    orbits = qv.concatenate([orbit, orbit, orbit])
    orbits = orbits.set_column(
        "orbit_id", pa.array(["orbit01", "orbit02", "orbit03"], type=pa.large_string())
    )

    fitted_orbits, fitted_orbits_members = evaluate_orbits(
        orbits, observations, propagator
    )

    # Check that the returned orbit is the same as the input orbit (this function
    # has merely evaluated the orbit, not changed it)
    assert fitted_orbits.orbit_id.to_pylist() == orbits.orbit_id.to_pylist()
    assert fitted_orbits.object_id.to_pylist() == orbits.object_id.to_pylist()
    assert fitted_orbits.coordinates == orbits.coordinates
    assert fitted_orbits.arc_length.to_pylist() == orbits.arc_length.to_pylist()
    assert fitted_orbits.num_obs.to_pylist() == orbits.num_obs.to_pylist()
    assert fitted_orbits.chi2.to_pylist() == orbits.chi2.to_pylist()
    assert fitted_orbits.reduced_chi2.to_pylist() == orbits.reduced_chi2.to_pylist()

    # Loop through each orbit and check that the returned orbit members are correctly evaluated
    for orbit_id in orbits.orbit_id.to_pylist():

        fitted_orbits_members_i = fitted_orbits_members.select("orbit_id", orbit_id)
        assert len(fitted_orbits_members_i) == len(orbit_members)
        assert fitted_orbits_members_i.orbit_id.tolist() == [
            orbit_id for _ in range(len(orbit_members))
        ]
        assert fitted_orbits_members_i.obs_id.tolist() == orbit_members.obs_id.tolist()
        assert (
            fitted_orbits_members_i.outlier.tolist() == orbit_members.outlier.tolist()
        )
        np.testing.assert_almost_equal(
            fitted_orbits_members_i.residuals.to_array(),
            orbit_members.residuals.to_array(),
        )


@pytest.mark.skipif(
    os.environ.get("OORB_DATA") is None, reason="OORB_DATA environment variable not set"
)
@pytest.mark.skipif(PYOORBPropagator is None, reason="PYOORBPropagator not available")
def test_evaluate_orbits_outliers(pure_iod_orbit):
    # Test that evaluate_orbit correctly calculates residuals and other
    # parameters for an input orbit with outliers defined
    orbit, orbit_members, observations = pure_iod_orbit
    propagator = PYOORBPropagator()

    # Lets remove the last two observations
    outliers = observations.id.tolist()[-2:]

    fitted_orbit, fitted_orbit_members = evaluate_orbits(
        orbit.to_orbits(), observations, propagator, ignore=outliers
    )

    # Check that the returned orbit's ID, object ID and coordinates are the same
    assert fitted_orbit.orbit_id[0].as_py() == orbit.orbit_id[0].as_py()
    assert fitted_orbit.object_id[0].as_py() == orbit.object_id[0].as_py()
    assert fitted_orbit.coordinates == orbit.coordinates

    # Because we marked two observations as outliers we expect that the arc length, number of observations
    # and chi2 values will be different
    assert fitted_orbit.arc_length[0].as_py() < orbit.arc_length[0].as_py()
    assert fitted_orbit.num_obs[0].as_py() == (orbit.num_obs[0].as_py() - 2)
    assert fitted_orbit.chi2[0].as_py() < orbit.chi2[0].as_py()
    assert fitted_orbit.reduced_chi2[0].as_py() < orbit.reduced_chi2[0].as_py()

    # Check that the returned orbit members are correctly evaluated
    assert len(fitted_orbit_members) == len(orbit_members)
    assert fitted_orbit_members.orbit_id.tolist() == orbit_members.orbit_id.tolist()
    assert fitted_orbit_members.obs_id.tolist() == orbit_members.obs_id.tolist()
    assert fitted_orbit_members.outlier.tolist() == [
        False,
        False,
        False,
        False,
        False,
        True,
        True,
    ]
    np.testing.assert_almost_equal(
        fitted_orbit_members.residuals.to_array(), orbit_members.residuals.to_array()
    )



from importlib.resources import files

import pytest

from ..differential_correction import OrbitDeterminationObservations
from ..fitted_orbits import FittedOrbitMembers, FittedOrbits


@pytest.fixture
def pure_iod_orbit():
    orbit = FittedOrbits.from_parquet(
        files("adam_core.orbit_determination.tests.data").joinpath(
            "pure_iod_orbit.parquet"
        )
    )
    orbit_members = FittedOrbitMembers.from_parquet(
        files("adam_core.orbit_determination.tests.data").joinpath(
            "pure_iod_orbit_members.parquet"
        )
    )
    observations = OrbitDeterminationObservations.from_parquet(
        files("adam_core.orbit_determination.tests.data").joinpath(
            "pure_iod_orbit_observations.parquet"
        )
    )
    return orbit, orbit_members, observations



from __future__ import annotations

import typing

from .cartesian import CartesianCoordinates
from .cometary import CometaryCoordinates
from .keplerian import KeplerianCoordinates
from .spherical import SphericalCoordinates

#: Type alias that represents a generic coordinate system.
Coordinates = typing.Union[
    CartesianCoordinates,
    KeplerianCoordinates,
    SphericalCoordinates,
    CometaryCoordinates,
]

#: Type variable which represents any of the coordinate classes.
CoordinateType = typing.TypeVar(
    "CoordinateType",
    bound=typing.Union[
        CartesianCoordinates,
        KeplerianCoordinates,
        CometaryCoordinates,
        SphericalCoordinates,
    ],
)



from importlib.resources import files
from typing import Optional

import numpy as np

from ...coordinates.covariances import CoordinateCovariances
from ...coordinates.keplerian import KeplerianCoordinates
from ...coordinates.origin import Origin
from ...orbits.orbits import Orbits
from ...time import Timestamp


def make_real_orbits(num_orbits: Optional[int] = None) -> Orbits:
    """
    Returns an `~adam_core.orbits.orbits.Orbits` object with real orbits drawn
    from our list of sample objects.

    Parameters
    ----------
    num_orbits : optional, int
        The number of orbits to return, which must be less than or equal to
        the number of sample objects (27).

    Returns
    -------
    orbits : `~adam_core.orbits.orbits.Orbits`
        Orbits object containing the sample orbits.
    """
    orbits_file = files("adam_core.utils.helpers.data").joinpath("orbits.parquet")
    orbits = Orbits.from_parquet(orbits_file)

    if num_orbits is None:
        return orbits

    if num_orbits > len(orbits):
        raise ValueError(
            f"num_orbits must be less than or equal to the number of sample orbits ({len(orbits)})."
        )

    return orbits[:num_orbits]


def make_simple_orbits(num_orbits: int = 10) -> Orbits:
    """
    Returns an `~adam_core.orbits.orbits.Orbits` object with simple orbits.

    Parameters
    ----------
    num_orbits : int, optional
        The number of orbits to return.
        Default is 10.

    Returns
    -------
    orbits : `~adam_core.orbits.orbits.Orbits`
        Orbits object containing the sample orbits.
    """

    data = {
        "a": np.linspace(1, 10, num_orbits),
        "e": np.linspace(0, 2, num_orbits),
        "i": np.linspace(0, 180, num_orbits),
        "raan": np.linspace(0, 360, num_orbits),
        "ap": np.linspace(0, 360, num_orbits),
        "M": np.linspace(0, 360, num_orbits),
    }
    # Hyperbolic orbits have negative semi-major axes
    data["a"] = np.where(data["e"] > 1.0, -data["a"], data["a"])

    sigmas = np.zeros((num_orbits, 6))
    for i, dim in enumerate(["a", "e", "i", "raan", "ap", "M"]):
        sigmas[:, i] = np.round(0.01 * data[dim], 4)

    data["covariance"] = CoordinateCovariances.from_sigmas(sigmas)
    data["time"] = Timestamp.from_mjd(
        np.round(np.linspace(59000.0, 59000.0 + num_orbits, num_orbits), 3),
        scale="tdb",
    )
    data["origin"] = Origin.from_kwargs(code=["SUN" for i in range(num_orbits)])
    data["frame"] = "ecliptic"

    coords = KeplerianCoordinates.from_kwargs(**data)
    object_ids = [f"Object {i:03d}" for i in range(num_orbits)]
    orbit_ids = [f"Orbit {i:03d}" for i in range(num_orbits)]

    orbits = Orbits.from_kwargs(
        coordinates=coords.to_cartesian(),
        orbit_id=orbit_ids,
        object_id=object_ids,
    )
    return orbits



from typing import Callable, Hashable, Optional, Tuple

import jax.numpy as jnp
import numpy as np
from jax import config, jacfwd, vmap

config.update("jax_enable_x64", True)
config.update("jax_platform_name", "cpu")

__all__ = ["calc_jacobian"]


def calc_jacobian(
    coords: np.ndarray,
    _func: Callable,
    in_axes: Optional[Hashable] = (0,),
    out_axes: Optional[int] = 0,
    **kwargs,
) -> jnp.ndarray:
    """
    Calculate the jacobian for the given callable in D dimensions for every
    N coordinate.

    Parameters
    ----------
    coords : `~numpy.ndarray` (N, D)
        Coordinates that correspond to the input covariance matrices.
    _func : function
        A function that takes a single coord (D) as input and return the transformed
        coordinate (D). See for example: `adam_core.coordinates.transform._cartesian_to_spherical`
        or `adam_core.coordinates.transform._cartesian_to_keplerian`.
    in_axes : Optional[Hashable]
        An integer or ``None`` indicates which array axis to map over for all arguments (with ``None``
        indicating not to map any axis), and a tuple indicates which axis to map
        for each corresponding positional argument. Axis integers must be in the
        range ``[-ndim, ndim)`` for each array, where ``ndim`` is the number of
        dimensions (axes) of the corresponding input array.

        From: https://jax.readthedocs.io/en/latest/_autosummary/jax.vmap.html#jax.vmap
    out_axes : Optional[int]
        An integer, None, or (nested) standard Python container (tuple/list/dict) thereof
        indicating where the mapped axis should appear in the output. All outputs with a
        mapped axis must have a non-None out_axes specification.

        From: https://jax.readthedocs.io/en/latest/_autosummary/jax.vmap.html#jax.vmap

    Returns
    -------
    jacobian : `~numpy.ndarray` (N, D, D)
        Array containing function partial derivatives for each coordinate.
    """
    # Calculate the jacobian function for the input function
    # Do this only once!
    jacobian_func = jacfwd(_func, argnums=0)

    vmapped_jacobian_func = vmap(
        jacobian_func,
        in_axes=in_axes,
        out_axes=out_axes,
    )

    jacobian = vmapped_jacobian_func(coords, *kwargs.values())
    # If the vmapped function returns more outputs, then only
    # return the first one. All relevant functions in adam_core return
    # primary result first, though we may want to come up with a more general
    # solution in the future.
    if isinstance(jacobian, Tuple):  # type: ignore
        jacobian = jacobian[0]
    return np.array(jacobian)



import importlib.util
import logging
from abc import abstractmethod
from typing import List, Optional, Tuple

import numpy as np
import numpy.typing as npt
import pyarrow.compute as pc
import quivr as qv

from adam_core.coordinates import CartesianCoordinates
from adam_core.ray_cluster import initialize_use_ray

from ..coordinates.residuals import Residuals
from ..orbits import Orbits
from ..orbits.variants import VariantOrbits
from ..propagator import Propagator
from ..propagator.propagator import OrbitType
from ..propagator.utils import _iterate_chunks
from ..time import Timestamp

logger = logging.getLogger(__name__)

# Test to see that at least one impact-enabled propagator is
# installed and if not print a warning
if importlib.util.find_spec("adam_core.propagator.adam_assist") is None:
    logger.warning(
        "No impact-enabled propagator installed. Impact calculations will not be possible."
    )

RAY_INSTALLED = False
try:
    import ray
    from ray import ObjectRef

    RAY_INSTALLED = True
except ImportError:
    pass


if RAY_INSTALLED:

    @ray.remote
    def impact_worker_ray(idx_chunk, orbits, propagator_class, num_days):
        prop = propagator_class()
        orbits_chunk = orbits.take(idx_chunk)
        variants, impacts = prop._detect_impacts(orbits_chunk, num_days)
        return variants, impacts


class EarthImpacts(qv.Table):
    orbit_id = qv.StringColumn()
    # Distance from earth center in km
    distance = qv.Float64Column()
    coordinates = CartesianCoordinates.as_column()
    variant_id = qv.LargeStringColumn(nullable=True)


class ImpactProbabilities(qv.Table):
    orbit_id = qv.LargeStringColumn()
    impacts = qv.Int64Column()
    variants = qv.Int64Column()
    cumulative_probability = qv.Float64Column()
    minimum_impact_time = Timestamp.as_column(nullable=True)
    maximum_impact_time = Timestamp.as_column(nullable=True)


class ImpactMixin:
    """
    `~adam_core.propagator.Propagator` mixin with signature for detecting Earth impacts.
    Subclasses should implement the _detect_impacts method.
    """

    @abstractmethod
    def _detect_impacts(
        self, orbits: Orbits, num_days: float
    ) -> Tuple[OrbitType, EarthImpacts]:
        """
        Detect impacts for the given orbits.

        THIS FUNCTION SHOULD BE DEFINED BY THE USER.
        """
        pass

    def detect_impacts(
        self,
        orbits: OrbitType,
        num_days: int,
        max_processes: Optional[int] = 1,
        chunk_size: int = 100,
    ) -> Tuple[OrbitType, EarthImpacts]:
        """
        Detect impacts for each orbit in orbits after num_days.

        Parameters
        ----------
        orbits : `~adam_core.orbits.orbits.Orbits` (N)
            Orbits for which to detect impacts.
        num_days : int
            Number of days after which to detect impacts.
        max_processes : int or None, optional
            Maximum number of processes to launch. If None then the number of
            processes will be equal to the number of cores on the machine. If 1
            then no multiprocessing will be used.

        Returns
        -------
        propagated : `~adam_core.orbits.OrbitType`
            The input orbits propagated to the end of simulation.
        impacts : `~adam_core.orbits.earth_impacts.EarthImpacts`
            Impacts detected for the orbits.
        """
        if max_processes is None or max_processes > 1:
            impact_list: List[EarthImpacts] = []
            propagated_list: List[OrbitType] = []

            if RAY_INSTALLED is False:
                raise ImportError(
                    "Ray must be installed to use the ray parallel backend"
                )

            initialize_use_ray(num_cpus=max_processes)

            # Add orbits to object store if
            # they haven't already been added
            if not isinstance(orbits, ObjectRef):
                orbits_ref = ray.put(orbits)
            else:
                orbits_ref = orbits
                # We need to dereference the orbits ObjectRef so we can
                # check its length for chunking and determine
                # if we need to propagate variants
                orbits = ray.get(orbits_ref)

            # Create futures
            futures = []
            idx = np.arange(0, len(orbits))
            for idx_chunk in _iterate_chunks(idx, chunk_size):
                futures.append(
                    impact_worker_ray.remote(
                        idx_chunk, orbits_ref, self.__class__, num_days
                    )
                )

            # Get results as they finish (we sort later)
            unfinished = futures
            while unfinished:
                finished, unfinished = ray.wait(unfinished, num_returns=1)
                (propagated, impacts) = ray.get(finished[0])
                propagated_list.append(propagated)
                impact_list.append(impacts)

            propagated = qv.concatenate(propagated_list)
            impacts = qv.concatenate(impact_list)

        else:
            propagated, impacts = self._detect_impacts(orbits, num_days)

        return propagated, impacts


def calculate_impacts(
    orbits: Orbits,
    num_days: int,
    propagator: Propagator,
    num_samples: int = 1000,
    processes: Optional[int] = None,
    seed: Optional[int] = None,
) -> Tuple[Orbits, EarthImpacts]:
    """
    Calculate the impacts for each variant orbit generated from the input orbits.

    Parameters
    ----------
    orbits : `~adam_core.orbits.orbits.Orbits`
        Orbits for which to calculate impact probabilities.
    num_days : int
        Number of days to propagate the orbits.
    propagator : `~adam_core.propagator.propagator.Propagator`
        Propagator to use for orbit propagation.
    num_samples : int, optional
        Number of samples to take over the period, by default 1000.
    processes : int, optional
        Number of processes to use for parallelization, by default all available.
    seed : int, optional
        Seed for random number generation, by default None.

    Returns
    -------

    """
    assert issubclass(
        propagator.__class__, ImpactMixin
    ), "Propagator must support impact detection."
    logger.info("Generating variants...")
    variants = VariantOrbits.create(
        orbits, method="monte-carlo", num_samples=num_samples, seed=seed
    )
    logger.info("Detecting impacts...")
    results, impacts = propagator.detect_impacts(
        variants,
        num_days,
        max_processes=processes,
    )

    return results, impacts


def calculate_impact_probabilities(
    variants: VariantOrbits, impacts: EarthImpacts
) -> ImpactProbabilities:
    """
    Calculate the impact probabilities for each variant orbit generated from the input orbits.
    Parameters
    ----------
    variants : `~adam_core.orbits.variants.VariantOrbits`
        Variant orbits for which to calculate impact probabilities.
    impacts : `~adam_core.orbits.impacts.Impacts`
        Impacts for the variant orbits.
    Returns
    -------
    impact_probabilities : `~adam_core.orbits.impact_probabilities.ImpactProbabilities`
        Impact probabilities for the variant orbits.
    """

    # Loop through the unique set of orbit_ids within variants using quivr
    unique_orbits = pc.unique(variants.orbit_id).to_pylist()

    earth_impact_probabilities = None

    for orbit_id in unique_orbits:
        # mask = pc.equal(variants.orbit_id, orbit_id)
        variant_masked = variants.select("orbit_id", orbit_id)
        variant_count = len(variant_masked)

        impacts_masked = impacts.select("orbit_id", orbit_id)
        impact_count = len(impacts_masked)

        ip = ImpactProbabilities.from_kwargs(
            orbit_id=[orbit_id],
            impacts=[impact_count],
            variants=[variant_count],
            cumulative_probability=[impact_count / variant_count],
        )

        if earth_impact_probabilities is None:
            earth_impact_probabilities = ip
        else:
            earth_impact_probabilities = qv.concatenate(
                [earth_impact_probabilities, ip]
            )

    return earth_impact_probabilities


def link_impacting_variants(variants, impacts):
    """
    Link variants to the orbits from which they were generated.
    Parameters
    ----------
    orbits : `~adam_core.orbits.orbits.Orbits`
        Orbits from which the variants were generated.
    Returns
    -------
    linkage : `~quivr.MultiKeyLinkage[Orbits, VariantOrbits]`
        Linkage between variants and orbits.
    """

    return qv.MultiKeyLinkage(
        impacts,
        variants,
        left_keys={
            "orbit_id": impacts.orbit_id,
            "variant_id": impacts.variant_id,
        },
        right_keys={
            "orbit_id": variants.orbit_id,
            "variant_id": variants.variant_id,
        },
    )


def calculate_mahalanobis_distance(
    observed_orbit: OrbitType, predicted_orbit: OrbitType
) -> npt.NDArray[np.float64]:
    """
    Calculate the Mahalanobis distance between an observed orbit and a predicted orbit.
    Parameters
    ----------
    observed_orbit : `~adam_core.coordinates.CartesianCoordinates`
        Observed orbit.
    observed_covariance : `ndarray`
        Covariance of the observed orbit.
    predicted_orbit : `~adam_core.coordinates.CartesianCoordinates`
        Predicted orbit.
    Returns
    -------
    mahalanobis_distance : `ndarray` (N)
        Mahalanobis distance between the observed and predicted orbits.
    """

    residuals = Residuals.calculate(
        observed_orbit.coordinates, predicted_orbit.coordinates
    )
    mahalanobis_distance = np.sqrt(residuals.chi2.to_numpy(zero_copy_only=False))
    return mahalanobis_distance



import numpy as np
import pyarrow as pa
import pytest
import quivr as qv

from ...coordinates.cartesian import CartesianCoordinates
from ...coordinates.origin import Origin
from ...observers.observers import Observers
from ...orbits.ephemeris import Ephemeris
from ...orbits.orbits import Orbits
from ...time.time import Timestamp
from ...utils.helpers.orbits import make_real_orbits
from ..propagator import EphemerisMixin, Propagator


class MockPropagator(Propagator, EphemerisMixin):
    # MockPropagator propagates orbits by just setting the time of the orbits.
    def _propagate_orbits(self, orbits: Orbits, times: Timestamp) -> Orbits:
        all_times = []
        for t in times:
            repeated_time = qv.concatenate([t] * len(orbits))
            orbits.coordinates.time = repeated_time
            all_times.append(orbits)

        return qv.concatenate(all_times)

    # MockPropagator generated ephemeris by just subtracting the state from
    # the state of the observers
    def _generate_ephemeris(self, orbits: Orbits, observers: Observers) -> Ephemeris:
        ephemeris_list = []
        observer_coordinates = observers.coordinates.values
        for orbit in orbits:
            topocentric_state = orbit.coordinates.values - observer_coordinates
            coords = CartesianCoordinates.from_kwargs(
                x=topocentric_state[:, 0],
                y=topocentric_state[:, 1],
                z=topocentric_state[:, 2],
                vx=topocentric_state[:, 3],
                vy=topocentric_state[:, 4],
                vz=topocentric_state[:, 5],
                time=observers.coordinates.time,
                origin=Origin.from_kwargs(code=observers.code),
                frame="ecliptic",
            )

            ephemeris_i = Ephemeris.from_kwargs(
                orbit_id=pa.array(np.full(len(coords), orbit.orbit_id[0].as_py())),
                object_id=pa.array(np.full(len(coords), orbit.object_id[0].as_py())),
                coordinates=coords.to_spherical(),
            )
            ephemeris_list.append(ephemeris_i)

        return qv.concatenate(ephemeris_list)


def test_propagator_single_worker():
    orbits = make_real_orbits(10)
    times = Timestamp.from_iso8601(["2020-01-01T00:00:00", "2020-01-01T00:00:01"])

    prop = MockPropagator()
    have = prop.propagate_orbits(orbits, times, max_processes=1)

    assert len(have) == len(orbits) * len(times)

    observers = Observers.from_code("X05", times)
    have = prop.generate_ephemeris(orbits, observers)

    assert len(have) == len(orbits) * len(times)


RAY_INSTALLED = False
try:
    import ray

    RAY_INSTALLED = True

except ImportError:
    pass


@pytest.mark.skipif(RAY_INSTALLED is False, reason="Ray is not installed.")
def test_propagator_multiple_workers_ray():
    orbits = make_real_orbits(10)
    times = Timestamp.from_iso8601(["2020-01-01T00:00:00", "2020-01-01T00:00:01"])

    prop = MockPropagator()
    have = prop.propagate_orbits(orbits, times, max_processes=4)

    assert len(have) == len(orbits) * len(times)

    observers = Observers.from_code("X05", times)
    have = prop.generate_ephemeris(orbits, observers, max_processes=4)

    assert len(have) == len(orbits) * len(times)

    # Now lets put times and orbits into the object store first
    orbits_ref = ray.put(orbits)
    time_ref = ray.put(times)

    have = prop.propagate_orbits(orbits_ref, time_ref, max_processes=4)

    assert len(have) == len(orbits) * len(times)

    observers_ref = ray.put(observers)
    have = prop.generate_ephemeris(orbits_ref, observers_ref, max_processes=4)

    assert len(have) == len(orbits) * len(times)



"""
Implementation of SGD-MOID

References
----------
[1] Hedo, J. M. et al. (2019). Minimum orbital intersection distance: an asymptotic approach.
    Astronomy & Astrophysics, 633, A22. https://doi.org/10.1051/0004-6361/201936502

Algorithm 1:
From an arbitrary point P_0 on the primary ellipse,
project the point onto the plane of the other ellipse.

Then find the minimum distance from the projected point,
using a minimization function.

Algorithm 2:

1. Divide up the primary ellipse into a series of points
2. For each point, run algorithm 1
3. After finding the point with the lowest distance,
search for the overall minimum.
"""

import numpy as np
import numpy.typing as npt
from scipy.optimize import minimize_scalar

from ..coordinates import CartesianCoordinates, KeplerianCoordinates
from ..dynamics.propagation import _propagate_2body
from ..orbits import Orbits
from ..time import Timestamp


def project_point_on_plane(
    P0: npt.NDArray, plane_coordinates: CartesianCoordinates
) -> npt.NDArray:
    """
    Take a point P0 (from the secondary ellipse) and project it onto the plane of the primary ellipse.
    """
    assert len(plane_coordinates) == 1

    n_hat = plane_coordinates.h[0] / plane_coordinates.h_mag[0]
    projected_point = P0 - np.dot(P0, n_hat) * n_hat
    return projected_point


def coplanar_distance_to_ellipse(
    P: npt.NDArray, keplerian_coordinates: KeplerianCoordinates, u: float
) -> float:
    """
    Calculate the distance from point P on the plane to the ellipse.
    """
    a, e, i, ap, raan, M = keplerian_coordinates.values[0]
    # Calculate the eccentric anomaly
    E = 2 * np.arctan(np.sqrt((1 - e) / (1 + e)) * np.tan(u / 2))
    # Calculate the true anomaly
    v = 2 * np.arctan(np.sqrt((1 + e) / (1 - e)) * np.tan(u / 2))
    # Calculate the radius
    r = a * (1 - e * np.cos(E))
    # Calculate the position vector
    r_vec = np.array([r * np.cos(v), r * np.sin(v), 0])
    # Calculate the distance from the point to the ellipse
    distance = np.linalg.norm(P - r_vec)
    return distance


def minimize_distance_from_coplanar_point_to_ellipse(
    P: npt.NDArray, keplerian_coordinates: KeplerianCoordinates
):
    """
    Find the minimum distance to the ellipse from planar point P

    We only need the magnitude of the distance here to calculate distance from P0 to closest point on ellipse
    """
    # Minimize the distance function with respect to an angle u
    result = minimize_scalar(
        lambda u: coplanar_distance_to_ellipse(P, keplerian_coordinates, u),
        bounds=(0, 2 * np.pi),
        method="bounded",
        tol=1e-12,
    )
    return result.fun


def distance_from_point_to_ellipse(
    P0: npt.NDArray, P: npt.NDArray, d_parallel: float
) -> float:
    """
    Takes the parallel and perpendicular distances from the point to the ellipse and calculates the total distance
    """
    d_perp = np.linalg.norm(P - P0)
    distance = np.sqrt(d_perp**2 + d_parallel**2)
    return distance


def calculate_distance_from_point_to_ellipse(P0: npt.NDArray, primary_ellipse: Orbits):
    """
    Calculate the distance from point P0 (from first ellipse) to the closest point on the ellipse
    """
    # Project the point onto the plane of the other ellipse
    P = project_point_on_plane(P0, primary_ellipse.coordinates)
    # Find the minimum distance from the projected point to the ellipse
    d_parallel = minimize_distance_from_coplanar_point_to_ellipse(
        P, primary_ellipse.coordinates.to_keplerian()
    )
    # Calculate the distance from P0 to the closest point on the ellipse
    distance = distance_from_point_to_ellipse(P0, P, d_parallel)
    return distance


def calculate_moid_for_dt(
    primary_ellipse: Orbits, secondary_ellipse: Orbits, dt: float
):
    # Propagate the primary ellipse by dt
    t_0 = primary_ellipse.coordinates.time.mjd()[0].as_py()
    primary_ellipse_propagated = _propagate_2body(
        primary_ellipse.coordinates.values[0],
        t_0,
        t_0 + dt,
        primary_ellipse.coordinates.origin.mu()[0],
    )
    P0 = primary_ellipse_propagated[:3]
    distance_to_secondary_ellipse = calculate_distance_from_point_to_ellipse(
        P0, secondary_ellipse
    )
    return distance_to_secondary_ellipse


# Now for Algorithm 1, where we discretize the primary ellipse and run Algorithm 1 for each point
def calculate_moid(
    primary_ellipse: Orbits, secondary_ellipse: Orbits
) -> tuple[float, Timestamp]:
    """
    Calculate the Minimum Orbit Intersection Distance (MOID) between two orbits.
    """
    keplerian = primary_ellipse.coordinates.to_keplerian()
    period = keplerian.P[0]
    e = keplerian.e[0].as_py()
    if e < 1:
        bounds = (0, period)
    else:
        bounds = (0, 10000)

    result = minimize_scalar(
        lambda dt: calculate_moid_for_dt(primary_ellipse, secondary_ellipse, dt),
        bounds=bounds,
        method="bounded",
        tol=1e-12,
    )
    moid = result.fun
    moid_time = Timestamp.from_mjd(
        [primary_ellipse.coordinates.time.mjd()[0].as_py() + result.x], scale="tdb"
    )
    if result.status != 0:
        raise ValueError("MOID calculation did not converge.")

    return moid, moid_time



import os
import pathlib

import numpy as np
import quivr as qv

from ...coordinates.cartesian import CartesianCoordinates
from ...time import Timestamp
from ..exposures import Exposures


def test_exposure_midpoints():
    start_times = Timestamp.from_iso8601(
        [
            "2000-01-01T00:00:00",
            "2000-01-02T00:00:00",
        ],
    )
    exp = Exposures.from_kwargs(
        id=["e1", "e2"],
        start_time=start_times,
        duration=[60, 30],
        filter=["g", "r"],
        observatory_code=["I41", "I41"],
    )

    midpoints = exp.midpoint()
    assert midpoints == Timestamp.from_iso8601(
        ["2000-01-01T00:00:30", "2000-01-02T00:00:15"]
    )


def test_exposure_states():
    observer_state_dir = (
        pathlib.Path(os.path.dirname(__file__))
        / ".."
        / ".."
        / "observers"
        / "tests"
        / "testdata"
    )

    w84_state_data = CartesianCoordinates.from_parquet(
        observer_state_dir / "W84_sun.parquet"
    )
    i41_state_data = CartesianCoordinates.from_parquet(
        observer_state_dir / "I41_sun.parquet"
    )

    # Mix up w84 and i41 in one big exposure table
    codes = ["W84", "I41", "W84", "I41", "W84"]
    state_times = qv.concatenate(
        [
            w84_state_data.time[0],
            i41_state_data.time[0],
            w84_state_data.time[3],
            i41_state_data.time[1],
            w84_state_data.time[2],
        ]
    )
    expected = qv.concatenate(
        [
            w84_state_data[0],
            i41_state_data[0],
            w84_state_data[3],
            i41_state_data[1],
            w84_state_data[2],
        ]
    )
    exp = Exposures.from_kwargs(
        id=["e1", "e2", "e3", "e4", "e5"],
        start_time=state_times,
        duration=[0, 0, 0, 0, 0],
        filter=["g", "r", "g", "r", "g"],
        observatory_code=codes,
    )

    obs = exp.observers()
    codes = obs.code

    assert codes.to_pylist() == ["W84", "I41", "W84", "I41", "W84"]
    states = obs.coordinates

    # Assert that states are same, to within expected number of bits
    # of precision.
    #
    # ideally, we'd like everything to be exactly equal to double
    # precision. That would require extremely careful numerical
    # analysis, though. The current values of maxulp are just
    # empirical.
    np.testing.assert_array_max_ulp(
        expected.x.to_numpy(), states.x.to_numpy(), dtype=np.float32
    )
    np.testing.assert_array_max_ulp(
        expected.y.to_numpy(), states.y.to_numpy(), dtype=np.float32
    )
    np.testing.assert_array_max_ulp(
        expected.z.to_numpy(), states.z.to_numpy(), maxulp=156, dtype=np.float32
    )
    np.testing.assert_array_max_ulp(
        expected.vx.to_numpy(), states.vx.to_numpy(), dtype=np.float32
    )
    np.testing.assert_array_max_ulp(
        expected.vy.to_numpy(), states.vy.to_numpy(), maxulp=2, dtype=np.float32
    )
    np.testing.assert_array_max_ulp(
        expected.vz.to_numpy(), states.vz.to_numpy(), maxulp=31, dtype=np.float32
    )

    np.testing.assert_allclose(expected.x.to_numpy(), states.x.to_numpy())
    np.testing.assert_allclose(expected.y.to_numpy(), states.y.to_numpy())
    np.testing.assert_allclose(expected.z.to_numpy(), states.z.to_numpy(), rtol=1e-4)
    np.testing.assert_allclose(expected.vx.to_numpy(), states.vx.to_numpy())
    np.testing.assert_allclose(expected.vy.to_numpy(), states.vy.to_numpy(), rtol=1e-6)
    np.testing.assert_allclose(expected.vz.to_numpy(), states.vz.to_numpy(), rtol=1e-5)


def test_empty_exposures():
    have = Exposures.empty()
    assert len(have) == 0



"""
This code generates the dictionary of semi-major axes for the
third body needed for the Tisserand parameter

from adam_core.orbits.query import _get_horizons_elements

ids = ["199", "299", "399", "499", "599", "699", "799", "899"]
elements = _get_horizons_elements(ids, times, id_type="majorbody")

MAJOR_BODIES = {}
for i, r in elements[["targetname", "a"]].iterrows():
   body_name = r["targetname"].split(" ")[0].lower()
   MAJOR_BODIES[body_name] = r["a"]

"""

import numpy as np

MAJOR_BODIES = {
    "mercury": 0.3870970330236769,
    "venus": 0.723341974974844,
    "earth": 0.9997889954736553,
    "mars": 1.523803685638066,
    "jupiter": 5.203719697535582,
    "saturn": 9.579110220472034,
    "uranus": 19.18646168457971,
    "neptune": 30.22486701698071,
}


def calc_tisserand_parameter(a, e, i, third_body="jupiter"):
    """
    Calculate Tisserand's parameter used to identify potential comets.
    For example, objects with Tisserand parameter's with respect to Jupiter greater than 3 are
    typically asteroids, whereas Jupiter family comets may have Tisserand's parameter
    between 2 and 3. Damocloids have Jupiter Tisserand's parameter of less than 2.

    Parameters
    ----------
    a : float or `~numpy.ndarray` (N)
        Semi-major axis in au.
    e : float or `~numpy.ndarray` (N)
        Eccentricity.
    i : float or `~numpy.ndarray` (N)
        Inclination in degrees.
    third_body : str
        Name of planet with respect to which Tisserand's parameter
        should be calculated.

    Returns
    -------
    Tp : float or `~numpy.ndarray` (N)
        Tisserand's parameter.
    """
    i_rad = np.radians(i)

    major_bodies = MAJOR_BODIES.keys()
    if third_body not in major_bodies:
        err = f"third_body should be one of {','.join(major_bodies)}"
        raise ValueError(err)

    ap = MAJOR_BODIES[third_body]
    Tp = ap / a + 2 * np.cos(i_rad) * np.sqrt(a / ap * (1 - e**2))

    return Tp
